

import tensorflow as tf

# from tensorflow.python.eager import context
# from tensorflow.python.framework import ops # tf.
# from tensorflow.python.ops import control_flow_ops
# from tensorflow.python.ops import math_ops
# from tensorflow.python.ops import resource_variable_ops
# from tensorflow.python.ops import state_ops
# from tensorflow.python.ops import variable_scope
# from tensorflow.python.training import optimizer
# from tensorflow.python.ops.clip_ops import clip_by_value

class AdaBoundOptimizer(tf.keras.optimizers.Optimizer):

    def __init__(self, learning_rate=0.001, final_lr=0.1, beta1=0.9, beta2=0.999,
                 gamma=1e-3, epsilon=1e-8, amsbound=False,
                 use_locking=False, name="AdaBound"):
        super(AdaBoundOptimizer, self).__init__(name)
        self._set_hyper("learning_rate", learning_rate)
        # self.lr = learning_rate
        self._final_lr = final_lr
        self._beta1 = beta1
        self._beta2 = beta2
        self._epsilon = epsilon

        self._gamma = gamma
        self._amsbound = amsbound

        self.lr_t = None
        self._beta1_t = None
        self._beta2_t = None
        self._epsilon_t = None

    def _create_slots(self, var_list):
        first_var = min(var_list, key=lambda x: x.name)

        graph = None if tf.executing_eagerly() else tf.get_default_graph()
        create_new = self._get_non_slot_variable("beta1_power", graph) is None
        if not create_new and tf.in_graph_mode():
            create_new = (self._get_non_slot_variable("beta1_power", graph).graph is not first_var.graph)

        if create_new:
            self._create_non_slot_variable(initial_value=self._beta1,
                                           name="beta1_power",
                                           colocate_with=first_var)
            self._create_non_slot_variable(initial_value=self._beta2,
                                           name="beta2_power",
                                           colocate_with=first_var)
            self._create_non_slot_variable(initial_value=self._gamma,
                                           name="gamma_multi",
                                           colocate_with=first_var)
        # Create slots for the first and second moments.
        for v in var_list :
            self._zeros_slot(v, "m", self._name)
            self._zeros_slot(v, "v", self._name)
            self._zeros_slot(v, "vhat", self._name)


    def _prepare(self):
        self.lr_t = tf.convert_to_tensor(self.lr)
        self._base_lr_t = tf.convert_to_tensor(self.lr)
        self._beta1_t = tf.convert_to_tensor(self._beta1)
        self._beta2_t = tf.convert_to_tensor(self._beta2)
        self._epsilon_t = tf.convert_to_tensor(self._epsilon)
        self._gamma_t = tf.convert_to_tensor(self._gamma)

    def _apply_dense(self, grad, var):
        graph = None if tf.executing_eagerly() else tf.get_default_graph()
        beta1_power = tf.cast(self._get_non_slot_variable("beta1_power", graph=graph), var.dtype.base_dtype)
        beta2_power = tf.cast(self._get_non_slot_variable("beta2_power", graph=graph), var.dtype.base_dtype)
        lr_t = tf.cast(self.lr_t, var.dtype.base_dtype)
        base_lr_t = tf.cast(self._base_lr_t, var.dtype.base_dtype)
        beta1_t = tf.cast(self._beta1_t, var.dtype.base_dtype)
        beta2_t = tf.cast(self._beta2_t, var.dtype.base_dtype)
        epsilon_t = tf.cast(self._epsilon_t, var.dtype.base_dtype)
        gamma_multi = tf.cast(self._get_non_slot_variable("gamma_multi", graph=graph), var.dtype.base_dtype)

        step_size = (lr_t * tf.sqrt(1 - beta2_power) / (1 - beta1_power))
        final_lr = self._final_lr * lr_t / base_lr_t
        lower_bound = final_lr * (1. - 1. / (gamma_multi + 1.))
        upper_bound = final_lr * (1. + 1. / (gamma_multi))

        # m_t = beta1 * m + (1 - beta1) * g_t
        m = self.get_slot(var, "m")
        m_scaled_g_values = grad * (1 - beta1_t)
        m_t = tf.assign(m, beta1_t * m + m_scaled_g_values, use_locking=self._use_locking)

        # v_t = beta2 * v + (1 - beta2) * (g_t * g_t)
        v = self.get_slot(var, "v")
        v_scaled_g_values = (grad * grad) * (1 - beta2_t)
        v_t = tf.assign(v, beta2_t * v + v_scaled_g_values, use_locking=self._use_locking)

        # amsgrad
        vhat = self.get_slot(var, "vhat")
        if self._amsbound :
            vhat_t = tf.assign(vhat, tf.maximum(v_t, vhat))
            v_sqrt = tf.sqrt(vhat_t)
        else :
            vhat_t = tf.assign(vhat, vhat)
            v_sqrt = tf.sqrt(v_t)


        # Compute the bounds
        step_size_bound = step_size / (v_sqrt + epsilon_t)
        bounded_lr = m_t * tf.clip_by_value(step_size_bound, lower_bound, upper_bound)

        var_update = tf.assign_sub(var, bounded_lr, use_locking=self._use_locking)
        return tf.group(*[var_update, m_t, v_t, vhat_t])

    def _resource_apply_dense(self, grad, var):
        graph = None if tf.executing_eagerly() else tf.get_default_graph()
        beta1_power = tf.cast(self._get_non_slot_variable("beta1_power", graph=graph), grad.dtype.base_dtype)
        beta2_power = tf.cast(self._get_non_slot_variable("beta2_power", graph=graph), grad.dtype.base_dtype)
        lr_t = tf.cast(self.lr_t, grad.dtype.base_dtype)
        base_lr_t = tf.cast(self._base_lr_t, var.dtype.base_dtype)
        beta1_t = tf.cast(self._beta1_t, grad.dtype.base_dtype)
        beta2_t = tf.cast(self._beta2_t, grad.dtype.base_dtype)
        epsilon_t = tf.cast(self._epsilon_t, grad.dtype.base_dtype)
        gamma_multi = tf.cast(self._get_non_slot_variable("gamma_multi", graph=graph), var.dtype.base_dtype)

        step_size = (lr_t * tf.sqrt(1 - beta2_power) / (1 - beta1_power))
        final_lr = self._final_lr * lr_t / base_lr_t
        lower_bound = final_lr * (1. - 1. / (gamma_multi + 1.))
        upper_bound = final_lr * (1. + 1. / (gamma_multi))

        # m_t = beta1 * m + (1 - beta1) * g_t
        m = self.get_slot(var, "m")
        m_scaled_g_values = grad * (1 - beta1_t)
        m_t = tf.assign(m, beta1_t * m + m_scaled_g_values, use_locking=self._use_locking)

        # v_t = beta2 * v + (1 - beta2) * (g_t * g_t)
        v = self.get_slot(var, "v")
        v_scaled_g_values = (grad * grad) * (1 - beta2_t)
        v_t = tf.assign(v, beta2_t * v + v_scaled_g_values, use_locking=self._use_locking)

        # amsgrad
        vhat = self.get_slot(var, "vhat")
        if self._amsbound:
            vhat_t = tf.assign(vhat, tf.maximum(v_t, vhat))
            v_sqrt = tf.sqrt(vhat_t)
        else:
            vhat_t = tf.assign(vhat, vhat)
            v_sqrt = tf.sqrt(v_t)

        # Compute the bounds
        step_size_bound = step_size / (v_sqrt + epsilon_t)
        bounded_lr = m_t * tf.clip_by_value(step_size_bound, lower_bound, upper_bound)

        var_update = tf.assign_sub(var, bounded_lr, use_locking=self._use_locking)

        return tf.group(*[var_update, m_t, v_t, vhat_t])

    def _apply_sparse_shared(self, grad, var, indices, scatter_add):
        graph = None if tf.executing_eagerly() else tf.get_default_graph()
        beta1_power = tf.cast(self._get_non_slot_variable("beta1_power", graph=graph), var.dtype.base_dtype)
        beta2_power = tf.cast(self._get_non_slot_variable("beta2_power", graph=graph), var.dtype.base_dtype)
        lr_t = tf.cast(self.lr_t, var.dtype.base_dtype)
        base_lr_t = tf.cast(self._base_lr_t, var.dtype.base_dtype)
        beta1_t = tf.cast(self._beta1_t, var.dtype.base_dtype)
        beta2_t = tf.cast(self._beta2_t, var.dtype.base_dtype)
        epsilon_t = tf.cast(self._epsilon_t, var.dtype.base_dtype)
        gamma_t = tf.cast(self._gamma_t, var.dtype.base_dtype)

        step_size = (lr_t * tf.sqrt(1 - beta2_power) / (1 - beta1_power))
        final_lr = self._final_lr * lr_t / base_lr_t
        lower_bound = final_lr * (1. - 1. / (gamma_t + 1.))
        upper_bound = final_lr * (1. + 1. / (gamma_t))

        # m_t = beta1 * m + (1 - beta1) * g_t
        m = self.get_slot(var, "m")
        m_scaled_g_values = grad * (1 - beta1_t)
        m_t = tf.assign(m, m * beta1_t, use_locking=self._use_locking)
        with tf.control_dependencies([m_t]):
            m_t = scatter_add(m, indices, m_scaled_g_values)

        # v_t = beta2 * v + (1 - beta2) * (g_t * g_t)
        v = self.get_slot(var, "v")
        v_scaled_g_values = (grad * grad) * (1 - beta2_t)
        v_t = tf.assign(v, v * beta2_t, use_locking=self._use_locking)
        with tf.control_dependencies([v_t]):
            v_t = scatter_add(v, indices, v_scaled_g_values)

        # amsgrad
        vhat = self.get_slot(var, "vhat")
        if self._amsbound:
            vhat_t = tf.assign(vhat, tf.maximum(v_t, vhat))
            v_sqrt = tf.sqrt(vhat_t)
        else:
            vhat_t = tf.assign(vhat, vhat)
            v_sqrt = tf.sqrt(v_t)

        # Compute the bounds
        step_size_bound = step_size / (v_sqrt + epsilon_t)
        bounded_lr = m_t * tf.clip_by_value(step_size_bound, lower_bound, upper_bound)

        var_update = tf.assign_sub(var, bounded_lr, use_locking=self._use_locking)

        return tf.group(*[var_update, m_t, v_t, vhat_t])

    def _apply_sparse(self, grad, var):
        return self._apply_sparse_shared(
            grad.values, var, grad.indices,
            lambda x, i, v: tf.scatter_add(  # pylint: disable=g-long-lambda
                x, i, v, use_locking=self._use_locking))

    def _resource_scatter_add(self, x, i, v):
        with tf.control_dependencies(
                [tf.scatter_add(x, i, v)]):
                # [resource_variable_ops.resource_scatter_add(x, i, v)]):
            return x.value()

    def _resource_apply_sparse(self, grad, var, indices):
        return self._apply_sparse_shared(
            grad, var, indices, self._resource_scatter_add)

    def _finish(self, update_ops, name_scope):
        # Update the power accumulators.
        with tf.control_dependencies(update_ops):
            graph = None if tf.executing_eagerly() else tf.get_default_graph()
            beta1_power = self._get_non_slot_variable("beta1_power", graph=graph)
            beta2_power = self._get_non_slot_variable("beta2_power", graph=graph)
            gamma_multi = self._get_non_slot_variable("gamma_multi", graph=graph)
            with tf.colocate_with(beta1_power):
                update_beta1 = beta1_power.assign(
                    beta1_power * self._beta1_t,
                    use_locking=self._use_locking)
                update_beta2 = beta2_power.assign(
                    beta2_power * self._beta2_t,
                    use_locking=self._use_locking)
                update_gamma = gamma_multi.assign(
                    gamma_multi + self._gamma_t,
                    use_locking=self._use_locking)
        return tf.group(*update_ops + [update_beta1, update_beta2, update_gamma],
                                      name=name_scope)


    def get_config(self):
        """Returns the config of the optimimizer.
        An optimizer config is a Python dictionary (serializable)
        containing the configuration of an optimizer.
        The same optimizer can be reinstantiated later
        (without any saved state) from this configuration.
        Returns:
            Python dictionary.
        """
        config = {"name": self._name}
        if hasattr(self, "lr"):
            config["learning_rate"] = self._serialize_hyperparameter("learning_rate")
        if hasattr(self, "_final_lr"):
            config["final_lr"] = self._final_lr
        if hasattr(self, "_beta1"):
            config["beta1"] = self._beta1
        if hasattr(self, "_beta2"):
            config["beta2"] = self._beta2
        if hasattr(self, "_gamma"):
            config["gamma"] = self._gamma
        if hasattr(self, "_asmbound"):
            config["asmbound"] = self._amsbound

        return config


"""
https://github.com/CyberZHG/keras-drop-block

MIT License

Copyright (c) 2018 Zhao HG

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
__all__ = ['DropBlock1D','DropBlock2D']

import tensorflow as tf

keras = tf.keras
K = tf.keras.backend

class DropBlock1D(keras.layers.Layer):
    """See: https://arxiv.org/pdf/1810.12890.pdf"""

    def __init__(self,
                 block_size,
                 keep_prob,
                 sync_channels=False,
                 data_format=None,
                 **kwargs):
        """Initialize the layer.
        :param block_size: Size for each mask block.
        :param keep_prob: Probability of keeping the original feature.
        :param sync_channels: Whether to use the same dropout for all channels.
        :param data_format: 'channels_first' or 'channels_last' (default).
        :param kwargs: Arguments for parent class.
        """
        super(DropBlock1D, self).__init__(**kwargs)
        self.block_size = block_size
        self.keep_prob = keep_prob
        self.sync_channels = sync_channels
        self.data_format = K.normalize_data_format(data_format)
        self.input_spec = keras.engine.base_layer.InputSpec(ndim=3)
        self.supports_masking = True

    def get_config(self):
        config = {'block_size': self.block_size,
                  'keep_prob': self.keep_prob,
                  'sync_channels': self.sync_channels,
                  'data_format': self.data_format}
        base_config = super(DropBlock1D, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_mask(self, inputs, mask=None):
        return mask

    def compute_output_shape(self, input_shape):
        return input_shape

    def _get_gamma(self, feature_dim):
        """Get the number of activation units to drop"""
        feature_dim = K.cast(feature_dim, K.floatx())
        block_size = K.constant(self.block_size, dtype=K.floatx())
        return ((1.0 - self.keep_prob) / block_size) * (feature_dim / (feature_dim - block_size + 1.0))

    def _compute_valid_seed_region(self, seq_length):
        positions = K.arange(seq_length)
        half_block_size = self.block_size // 2
        valid_seed_region = K.switch(
            K.all(
                K.stack(
                    [
                        positions >= half_block_size,
                        positions < seq_length - half_block_size,
                    ],
                    axis=-1,
                ),
                axis=-1,
            ),
            K.ones((seq_length,)),
            K.zeros((seq_length,)),
        )
        return K.expand_dims(K.expand_dims(valid_seed_region, axis=0), axis=-1)

    def _compute_drop_mask(self, shape):
        seq_length = shape[1]
        mask = K.random_binomial(shape, p=self._get_gamma(seq_length))
        mask *= self._compute_valid_seed_region(seq_length)
        mask = keras.layers.MaxPool1D(
            pool_size=self.block_size,
            padding='same',
            strides=1,
            data_format='channels_last',
        )(mask)
        return 1.0 - mask

    def call(self, inputs, training=None):

        def dropped_inputs():
            outputs = inputs
            if self.data_format == 'channels_first':
                outputs = K.permute_dimensions(outputs, [0, 2, 1])
            shape = K.shape(outputs)
            if self.sync_channels:
                mask = self._compute_drop_mask([shape[0], shape[1], 1])
            else:
                mask = self._compute_drop_mask(shape)
            outputs = outputs * mask *\
                (K.cast(K.prod(shape), dtype=K.floatx()) / K.sum(mask))
            if self.data_format == 'channels_first':
                outputs = K.permute_dimensions(outputs, [0, 2, 1])
            return outputs

        return K.in_train_phase(dropped_inputs, inputs, training=training)


class DropBlock2D(keras.layers.Layer):
    """See: https://arxiv.org/pdf/1810.12890.pdf"""

    def __init__(self,
                 block_size,
                 keep_prob,
                 sync_channels=False,
                 data_format=None,
                 **kwargs):
        """Initialize the layer.
        :param block_size: Size for each mask block.
        :param keep_prob: Probability of keeping the original feature.
        :param sync_channels: Whether to use the same dropout for all channels.
        :param data_format: 'channels_first' or 'channels_last' (default).
        :param kwargs: Arguments for parent class.
        """
        super(DropBlock2D, self).__init__(**kwargs)
        self.block_size = block_size
        self.keep_prob = keep_prob
        self.sync_channels = sync_channels
        self.data_format = K.normalize_data_format(data_format)
        self.input_spec = keras.engine.base_layer.InputSpec(ndim=4)
        self.supports_masking = True

    def get_config(self):
        config = {'block_size': self.block_size,
                  'keep_prob': self.keep_prob,
                  'sync_channels': self.sync_channels,
                  'data_format': self.data_format}
        base_config = super(DropBlock2D, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_mask(self, inputs, mask=None):
        return mask

    def compute_output_shape(self, input_shape):
        return input_shape

    def _get_gamma(self, height, width):
        """Get the number of activation units to drop"""
        height, width = K.cast(height, K.floatx()), K.cast(width, K.floatx())
        block_size = K.constant(self.block_size, dtype=K.floatx())
        return ((1.0 - self.keep_prob) / (block_size ** 2)) *\
               (height * width / ((height - block_size + 1.0) * (width - block_size + 1.0)))

    def _compute_valid_seed_region(self, height, width):
        positions = K.concatenate([
            K.expand_dims(K.tile(K.expand_dims(K.arange(height), axis=1), [1, width]), axis=-1),
            K.expand_dims(K.tile(K.expand_dims(K.arange(width), axis=0), [height, 1]), axis=-1),
        ], axis=-1)
        half_block_size = self.block_size // 2
        valid_seed_region = K.switch(
            K.all(
                K.stack(
                    [
                        positions[:, :, 0] >= half_block_size,
                        positions[:, :, 1] >= half_block_size,
                        positions[:, :, 0] < height - half_block_size,
                        positions[:, :, 1] < width - half_block_size,
                    ],
                    axis=-1,
                ),
                axis=-1,
            ),
            K.ones((height, width)),
            K.zeros((height, width)),
        )
        return K.expand_dims(K.expand_dims(valid_seed_region, axis=0), axis=-1)

    def _compute_drop_mask(self, shape):
        height, width = shape[1], shape[2]
        mask = K.random_binomial(shape, p=self._get_gamma(height, width))
        mask *= self._compute_valid_seed_region(height, width)
        mask = keras.layers.MaxPool2D(
            pool_size=(self.block_size, self.block_size),
            padding='same',
            strides=1,
            data_format='channels_last',
        )(mask)
        return 1.0 - mask

    def call(self, inputs, training=None):

        def dropped_inputs():
            outputs = inputs
            if self.data_format == 'channels_first':
                outputs = K.permute_dimensions(outputs, [0, 2, 3, 1])
            shape = K.shape(outputs)
            if self.sync_channels:
                mask = self._compute_drop_mask([shape[0], shape[1], shape[2], 1])
            else:
                mask = self._compute_drop_mask(shape)
            outputs = outputs * mask *\
                (K.cast(K.prod(shape), dtype=K.floatx()) / K.sum(mask))
            if self.data_format == 'channels_first':
                outputs = K.permute_dimensions(outputs, [0, 3, 1, 2])
            return outputs

        return K.in_train_phase(dropped_inputs, inputs, training=training)


import os

import distutils
from distutils import dir_util

import tensorflow as tf

from . context import Context

COLAB_PATH = "/content/drive/My Drive"
# COLAB_PATH = "./tmp/colab"
COLAB_BACKUP_BASE_GDRIVE = COLAB_PATH + "/tftk/colab_back"

def IS_ON_COLABOLATORY_WITH_GOOGLE_DRIVE()->bool:
    return tf.io.gfile.exists(COLAB_PATH)

class Colaboratory():

    resumed = False

    @classmethod
    def copy_resume_data_from_google_drive(cls):
        context = Context.get_instance()
        name = context[Context.TRAINING_NAME]
        base = context[Context.TRAINING_BASE_DIR]

        colab = base + os.path.sep + name
        gdrive = COLAB_BACKUP_BASE_GDRIVE + os.path.sep + name

        if tf.io.gfile.exists(gdrive)==True and cls.resumed == False:
            print("copy files from gdrive")
            distutils.dir_util.copy_tree(gdrive, colab,verbose=1)
            cls.resume = True
        else:
            print("No resume data on gdrive.")

    @classmethod
    def copy_suspend_data_from_colab(cls):
        context = Context.get_instance()
        name = context[Context.TRAINING_NAME]
        base = context[Context.TRAINING_BASE_DIR]

        colab = base + os.path.sep + name
        gdrive = COLAB_BACKUP_BASE_GDRIVE + os.path.sep + name

        if tf.io.gfile.exists(gdrive) == False:
            print("google drive dirctory not found ", gdrive)
            os.makedirs(gdrive)
        else:
            print("gdrive found ", gdrive)

        print("\nbackup suspend data ", colab, gdrive,"\n")
        distutils.dir_util.copy_tree(colab,gdrive,verbose=1)

            



import tensorflow as tf
import tftk

class OptimizerBuilder():

    @classmethod
    def get_optimizer(cls,name:str,**kwargs)->tf.keras.optimizers.Optimizer:
        """Optimizerを取得する。
        
        Arguments:
            name {str} --　オプティマイザの名称{sgd/adadelta}

            lr {float}  -- SGDの学習率(デフォルト 0.01) , adadeltaの学習率(デフォルト 1.0)
            sgd_momentum {float} -- SGDのモーメンタム,デフォルト 0.0
            sgd_nestrov {bool} -- SGDのnestrov
            adadelta_rho {float} -- デフォルト 0.95
            adadelta_epsilon {float} -- adadeltaのepsilon デフォルトNone
            adadelta_decay {float} -- adadeltaのdecay デフォルト0.0
        
        Returns:
            tf.keras.optimizers.Optimizer -- [description]
        """
        ret = None
        if name == "sgd":
            learning_rate = kwargs.get("lr", 0.01)
            momentum = kwargs.get("sgd_momentum", 0.0)
            nesterov = kwargs.get("sgd_nestrov" , False)
            print("learning_rate",learning_rate,"momentum",momentum,"nestrov",nesterov)
            ret = tf.keras.optimizers.SGD(learning_rate=learning_rate,momentum=momentum,nesterov=nesterov)
        elif name == "adadelta":
            # lr=1.0, rho=0.95, epsilon=None, decay=0.0
            lr = kwargs.get("lr", 1.0)
            rho = kwargs.get("adadelta_rho", 0.95)
            epsilon = kwargs.get("adadelta_epsilon",None)
            decay = kwargs.get("adadelta_decay",0.0)
            print("lr",lr,"rho", rho , "epsiloon", epsilon, "decay",decay)
            ret = tf.keras.optimizers.Adadelta(lr=lr,rho=rho,epsilon=epsilon,decay=decay)
        elif name=="adam":
            # learning_rate = kwargs.get("lr",0.001)
            # beta_1 = kwargs.get("beta_1",0.9)
            # beta_2 = kwargs.get("beta_2",0.999)
            # epsilon = kwargs.get("epsilon",1e-8)
            # ret = tf.keras.optimizers.Adam(learning_rate=learning_rate,beta_1=beta_1,beta_2=beta_2,epsilon=epsilon)
            ret = tf.keras.optimizers.Adam()
        elif name =='rmsprop':
            lr = kwargs.get("lr", 0.001)
            ret = tf.keras.optimizers.RMSprop(lr=lr)
        
        if tftk.IS_MIXED_PRECISION() == True:
            ret = tf.keras.mixed_precision.experimental.LossScaleOptimizer(ret, loss_scale='dynamic')       

        return ret


    


import os
import glob
from typing import List,Dict
import shutil

import tensorflow as tf
import tensorflow_datasets as tfds
from icrawler.builtin import BingImageCrawler

from tftk.image.dataset import BaseDataset

class Mnist(BaseDataset):
    @classmethod
    def get_train_dataset(cls,**kwargs)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="mnist", split="train",with_info=True)
        return ds, info.splits["train"].num_examples

    @classmethod
    def get_validation_dataset(cls, **kwargs)->(tf.data.Dataset, int):
        return (None, -1)


    @classmethod
    def get_test_dataset(cls,**kwargs)->(tf.data.Dataset, int):
        ds, info =  tfds.load(name="mnist", split="test",with_info=True)
        return ds, info.splits["test"].num_examples


class Cifar10(BaseDataset):

    @classmethod
    def get_train_dataset(cls)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="cifar10", split="train",with_info=True)
        return ds, info.splits["train"].num_examples

    @classmethod
    def get_validation_dataset(cls, **kwargs)->(tf.data.Dataset, int):
        return (None, -1)

    @classmethod
    def get_test_dataset(cls)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="cifar10", split="test",with_info=True)
        return ds, info.splits["test"].num_examples



class ImageCrawler():

    @classmethod
    def crawl_keywords_save_folder(cls, name:str, keywords:List[str], base_dir:str="tmp", filters:Dict={"size":"large"}, max_num:int= 10000, train_ratio:float=0.8):
        """キーワードでクロールしてデータセットを作成する。
            Arguments:
            name {str} -- データセット名
            keywords {List[str]} -- キーワード
            filter {[type]} -- [description]

        Keyword Arguments:
            dest_base_dir {str} -- [description] (default: {"tmp/crawl"})
            train_ratio {float} -- [description] (default: {0.8})
        """
        download_base = base_dir + os.path.sep + name
        for k in keywords:
            download_dir = download_base + os.path.sep + "train" + os.path.sep + k
            if os.path.exists(download_dir) != True:
                os.makedirs(download_dir)
            storage={"root_dir": download_dir }

            print( "keyword:",k, " dir",download_dir)

            crawler = BingImageCrawler(storage=storage)
            crawler.crawl(keyword=k, filters=filters, max_num=max_num, file_idx_offset=0)


            move_dir = download_base + os.path.sep + "test" + os.path.sep + k
            if os.path.exists(move_dir) != True:
                os.makedirs(move_dir)

            file_list = glob.glob(download_dir+os.path.sep+"*.jpg")
            move_num:int = int(len(file_list) * train_ratio)

            move_list = file_list[move_num:]
            for f in move_list:
                shutil.move(f,move_dir)




class ImageLabelFolderDataset():

    @classmethod
    def get_train_dataset(cls, name="dataset", manual_dir="./", **kwargs)->(tf.data.Dataset, int):
        print(name,manual_dir)
        builder = tfds.image.ImageLabelFolder(name)
        dl_config = tfds.download.DownloadConfig(manual_dir=manual_dir)
        builder.download_and_prepare(download_config=dl_config)
        ds = builder.as_dataset(split='train', shuffle_files=False)
        len = builder.info.splits['train'].num_examples  # Splits, num examples,... automatically extracted
        return ds, len


    @classmethod
    def get_validation_dataset(cls, name="dataset", dir="./", **kwargs)->(tf.data.Dataset,int):
        """validation用データセットを取得する。
        
        Arguments:
            int {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        builder = tfds.image.ImageLabelFolder(name)
        dl_config = tfds.download.DownloadConfig(manual_dir=dir)
        builder.download_and_prepare(download_config=dl_config)
        ds = builder.as_dataset(split='validation', shuffle_files=False)
        len = builder.info.splits['validation'].num_examples  # Splits, num examples,... automatically extracted
        return ds, len

    @classmethod
    def get_test_dataset(cls, name="dataset", dir="./", **kwargs)->(tf.data.Dataset,int):
        """test用データセットを取得する。
        
        Arguments:
            int {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        builder = tfds.image.ImageLabelFolder(name)
        dl_config = tfds.download.DownloadConfig(manual_dir=dir)
        builder.download_and_prepare(download_config=dl_config)
        ds = builder.as_dataset(split='test', shuffle_files=False)
        len = builder.info.splits['test'].num_examples  # Splits, num examples,... automatically extracted
        return ds, len


class Place365Small(BaseDataset):
    @classmethod
    def get_train_dataset(cls,**kwargs)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="places365_small", split="train",with_info=True)
        return ds, info.splits["train"].num_examples
    
    @classmethod
    def get_validation_dataset(cls,**kwargs)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="places365_small", split="validation",with_info=True)
        return ds, info.splits["validation"].num_examples

    @classmethod
    def get_test_dataset(cls,**kwargs)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="places365_small", split="test",with_info=True)
        return ds, info.splits["test"].num_examples




class Food1o1(BaseDataset):
    @classmethod
    def get_train_dataset(cls,**kwargs)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="food101", split="train",with_info=True)
        return ds, info.splits["train"].num_examples

    @classmethod
    def get_validation_dataset(cls,**kwargs)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="food101", split="validation",with_info=True)
        return ds, info.splits["validation"].num_examples

    @classmethod
    def get_test_dataset(cls,**kwargs)->(tf.data.Dataset, int):
        return (None,-1)

class PatchCamelyon(BaseDataset):
    @classmethod
    def get_train_dataset(cls,**kwargs)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="patch_camelyon", split="train",with_info=True)
        return ds, info.splits["train"].num_examples

    @classmethod
    def get_validation_dataset(cls,**kwargs)->(tf.data.Dataset, int):
        return (None,-1)

    @classmethod
    def get_test_dataset(cls,**kwargs)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="patch_camelyon", split="test",with_info=True)
        return ds, info.splits["test"].num_examples

class ImageNetResized(BaseDataset):
    """ImageNetリサイズ画像
    
    Arguments:   
        BaseDataset {[type]} -- [description]
    
    Returns:
        [type] -- [description]
    """
    @classmethod
    def get_train_dataset(cls,**kwargs)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="imagenet_resized/64x64", split="train",with_info=True)
        return ds, info.splits["train"].num_examples

    @classmethod
    def get_test_dataset(cls,**kwargs)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="imagenet_resized/64x64", split="validation",with_info=True)
        return ds, info.splits["validation"].num_examples

    @classmethod
    def get_test_dataset(cls,**kwargs)->(tf.data.Dataset, int):
        return (None,-1)


# bellows are working...

class ImageNet2012(BaseDataset):
    @classmethod
    def get_train_dataset(cls)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="imagenet2012", split="train",with_info=True,data_dir="D:\\imagenet")
        return ds, 1281167

    @classmethod
    def get_test_dataset(cls)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="imagenet2012", split="ttest",with_info=True,data_dir="D:\\imagenet")
        return ds, info.splits["test"].num_examples


class CatsVsDogs(BaseDataset):
    @classmethod
    def get_train_dataset(cls)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="cats_vs_dogs", split="train",with_info=True)
        return ds, info.splits["train"].num_examples

    @classmethod
    def get_test_dataset(cls)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="cats_vs_dogs", split="test",with_info=True)
        return ds, info.splits["test"].num_examples

class RockPaperScissors(BaseDataset):

    @classmethod
    def get_train_dataset(cls)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="rock_paper_scissors", split="train",with_info=True)
        return ds, info.splits["train"].num_examples

    @classmethod
    def get_test_dataset(cls)->(tf.data.Dataset, int):
        ds, info = tfds.load(name="rock_paper_scissors", split="test",with_info=True)
        return ds, info.splits["test"].num_examples



import math

import tensorflow as tf
import tensorflow_probability as tfp
import tensorflow_addons as tfa

from typing import Callable
from typing import Dict

class ImageAugument():

    _MAX_LEVEL = 10.

    @classmethod
    def mixup_apply(cls, mixup_size: int, alpha: float) -> Callable[[tf.data.Dataset], tf.data.Dataset]:
        """Mixup拡張をおこなう。ラベルはone-hot化されている事。
           dataset:tf.data.Dataset = somedataset
           mixuped = dataset.apply(DataAugument.mixup_apply(100,0.8))

        Args:
            dataset (tf.data.Dataset): dataset to be applied mixup augmentation.
            mix_size (int): data shuffle size to mixup
            alpha (float): Hyper parameter of the mixup. (0-1)

        Returns:
            tf.data.Dataset: mixuped dataset.
        """

        def mixup(dataset: tf.data.Dataset) -> tf.data.Dataset:
            def mixup_map(data, shuffled):
                dist = tfp.distributions.Beta([alpha], [alpha])
                beta = dist.sample([1])[0][0]

                ret = {}
                ret["image"] = (data["image"]) * beta + (shuffled["image"] * (1-beta))
                ret["label"] = (data["label"]) * beta + (shuffled["label"] * (1-beta))
                return ret

            shuffle_dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE).shuffle(mixup_size)
            zipped = tf.data.Dataset.zip((dataset, shuffle_dataset))
            return zipped.map(mixup_map, num_parallel_calls=tf.data.experimental.AUTOTUNE)

        return mixup

    # from tensorflow/tpu autoaugment.py

    @classmethod
    def _blend(cls, image1, image2, factor):
        """_blend image1 and image2 using 'factor'.
        Factor can be above 0.0.  A value of 0.0 means only image1 is used.
        A value of 1.0 means only image2 is used.  A value between 0.0 and
        1.0 means we linearly interpolate the pixel values between the two
        images.  A value greater than 1.0 "extrapolates" the difference
        between the two pixel values, and we clip the results to values
        between 0 and 255.

        Args:
            image1: An image Tensor of type uint8.
            image2: An image Tensor of type uint8.
            factor: A floating point value above 0.0.
        
        Returns:
            A _blended image Tensor of type uint8.
        """
        if factor == 0.0:
            return tf.convert_to_tensor(image1)
        if factor == 1.0:
            return tf.convert_to_tensor(image2)

        image1 = tf.cast(image1, tf.float32)
        image2 = tf.cast(image2, tf.float32)

        difference = image2 - image1
        scaled = factor * difference

        # Do addition in float.
        temp = tf.cast(image1, tf.float32) + scaled

        # Interpolate
        if factor > 0.0 and factor < 1.0:
        # Interpolation means we always stay within 0 and 255.
            return tf.cast(temp, tf.uint8)

        # Extrapolate:
        #
        # We need to clip and then cast.
        return tf.cast(tf.clip_by_value(temp, 0.0, 255.0), tf.uint8)

    @classmethod
    def cutout_map(cls, pad_size):
        """cutoutのみ単体で拡張する場合のmap用関数
        Args:
            pad_size: カットするサイズ 縦横ともに(pad_size*2)pxを切り取り 
        Returns:
            カットアウト関数
        """

        def cutout(data):
            data["image"] = cls._cutout(data["image"], pad_size)
            return data

        return cutout

    @classmethod
    def cutout(cls,image, pad_size, replace=0):
        """Apply cutout (https://arxiv.org/abs/1708.04552) to image.
        This operation applies a (2*pad_size x 2*pad_size) mask of zeros to
        a random location within `img`. The pixel values filled in will be of the
        value `replace`. The located where the mask will be applied is randomly
        chosen uniformly over the whole image.
        Args:
            image: An image Tensor of type uint8.
            pad_size: Specifies how big the zero mask that will be generated is that
            is applied to the image. The mask will be of size
            (2*pad_size x 2*pad_size).
            replace: What pixel value to fill in the image in the area that has
            the cutout mask applied to it.
        Returns:
            An image Tensor that is of type uint8.
        """
        image_height = tf.shape(image)[0]
        image_width = tf.shape(image)[1]

        # Sample the center location in the image where the zero mask will be applied.
        cutout_center_height = tf.random.uniform(
            shape=[], minval=0, maxval=image_height,
            dtype=tf.int32)
        
        cutout_center_width = tf.random.uniform(
            shape=[], minval=0, maxval=image_width,
            dtype=tf.int32)

        lower_pad = tf.math.maximum(0, cutout_center_height - pad_size)
        upper_pad = tf.math.maximum(0, image_height - cutout_center_height - pad_size)
        left_pad = tf.math.maximum(0, cutout_center_width - pad_size)
        right_pad = tf.math.maximum(0, image_width - cutout_center_width - pad_size)

        cutout_shape = [image_height - (lower_pad + upper_pad),
                  image_width - (left_pad + right_pad)]
        padding_dims = [[lower_pad, upper_pad], [left_pad, right_pad]]
        mask = tf.pad(
            tf.zeros(cutout_shape, dtype=image.dtype),
            padding_dims, constant_values=1)
        mask = tf.expand_dims(mask, -1)
        mask = tf.tile(mask, [1, 1, 3])
        image = tf.where(
            tf.equal(mask, 0),
            tf.ones_like(image, dtype=image.dtype) * replace,
            image)
        return image


    @classmethod 
    def solarize(cls, image, threshold=128):
        # For each pixel in the image, select the pixel
        # if the value is less than the threshold.
        # Otherwise, subtract 255 from the pixel.
        return tf.where(image < threshold, image, 255 - image)

    @classmethod
    def solarize_add(cls, image, additionimage, addition=0, threshold=128):
        # For each pixel in the image less than threshold
        # we add 'addition' amount to it and then clip the
        # pixel value to be between 0 and 255. The value
        # of 'addition' is between -128 and 128.
        added_image = tf.cast(image, tf.int64) + addition
        added_image = tf.cast(tf.clip_by_value(added_image, 0, 255), tf.uint8)
        return tf.where(image < threshold, added_image, image)

    @classmethod
    def color(cls, image, factor):
        """Equivalent of PIL Color."""
        degenerate = tf.image.grayscale_to_rgb(tf.image.rgb_to_grayscale(image))
        return cls._blend(degenerate, image, factor)
    
    @classmethod
    def contrast(cls, image, factor):
        """Equivalent of PIL Contrast."""
        degenerate = tf.image.rgb_to_grayscale(image)
        # Cast before calling tf.histogram.
        degenerate = tf.cast(degenerate, tf.int32)

        # Compute the grayscale histogram, then compute the mean pixel value,
        # and create a constant image size of that value.  Use that as the
        # _blending degenerate target of the original image.
        hist = tf.histogram_fixed_width(degenerate, [0, 255], nbins=256)
        mean = tf.reduce_sum(tf.cast(hist, tf.float32)) / 256.0
        degenerate = tf.ones_like(degenerate, dtype=tf.float32) * mean
        degenerate = tf.clip_by_value(degenerate, 0.0, 255.0)
        degenerate = tf.image.grayscale_to_rgb(tf.cast(degenerate, tf.uint8))
        return cls._blend(degenerate, image, factor)

    @classmethod
    def brightness(cls, image, factor):
        """Equivalent of PIL Brightness."""
        degenerate = tf.zeros_like(image)
        return cls._blend(degenerate, image, factor)

    @classmethod
    def posterize(cls, image, bits):
        """Equivalent of PIL Posterize."""
        shift = 8 - bits
        return tf.bitwise.left_shift(tf.bitwise.right_shift(image, shift), shift)
    
    @classmethod
    def rotate(cls, image, degrees):
        """Rotates the image by degrees either clockwise or counterclockwise.
        Args:
            image: An image Tensor of type uint8.
            degrees: Float, a scalar angle in degrees to rotate all images by. If
            degrees is positive the image will be rotated clockwise otherwise it will
            be rotated counterclockwise.
        Returns:
            The rotated version of image.
        """
        # Convert from degrees to radians.
        print(image)
        degrees_to_radians = math.pi / 180.0
        radians = degrees * degrees_to_radians

        rotate_image = tfa.image.rotate(image, radians,name="rotate")
        # image = contrib_image.rotate(_warp(image), radians)
        return rotate_image

    @classmethod
    def translate_x(cls, image, pixels, replace=[128,128,128]):
        """Equivalent of PIL Translate in X dimension."""
        image = tfa.image.translate(image,[-pixels,0])
        # image = contrib_image.translate(_warp(image), [-pixels, 0])
        return image # un_warp(image, replace)

    @classmethod
    def translate_y(cls, image, pixels, replace=[128,128,128]):
        """Equivalent of PIL Translate in Y dimension."""
        image = tfa.image.translate(image,[0,-pixels])
        return image

    @classmethod
    def shear_x(cls, image, level, replace=[128,128,128]):
        """Equivalent of PIL Shearing in X dimension."""
        # Shear parallel to x axis is a projective transform
        # # with a matrix form of:
        # [1  level
        #  0  1].
        image = tfa.image.transform(
            image, [1., level, 0., 0., 1., 0., 0., 0.])
        return image

    @classmethod
    def shear_y(cls, image, level, replace=[128,128,128]):
        """Equivalent of PIL Shearing in Y dimension."""
        # Shear parallel to y axis is a projective transform
        # with a matrix form of:
        # [1  0
        #  level  1].
        image = tfa.image.transform(
            image, [1., 0., 0., level, 1., 0., 0., 0.])
        return image

    @classmethod
    def autocontrast(cls, image):
        """Implements Autocontrast function from PIL using TF ops.
            Args:
                image: A 3D uint8 tensor.
            Returns:
                The image after it has had autocontrast applied to it and will be of type
                uint8.
        """
        def scale_channel(image):
            """Scale the 2D image using the autocontrast rule."""
            # A possibly cheaper version can be done using cumsum/unique_with_counts
            # over the histogram values, rather than iterating over the entire image.
            # to compute mins and maxes.

            lo = tf.cast(tf.reduce_min(image),tf.float32)
            hi = tf.cast(tf.reduce_max(image),tf.float32)

            # Scale the image, making the lowest value 0 and the highest value 255.
            def scale_values(im):
                scale = 255.0 / (hi - lo)
                offset = -lo * scale
                im = tf.cast(im,tf.float32) * scale + offset
                im = tf.clip_by_value(im, 0.0, 255.0)
                return tf.cast(im, tf.uint8)

            result = tf.cond(hi > lo, lambda: scale_values(image), lambda: image)
            return result

        # Assumes RGB for now.  Scales each channel independently
        # and then stacks the result.
        s1 = scale_channel(image[:, :, 0])
        s2 = scale_channel(image[:, :, 1])
        s3 = scale_channel(image[:, :, 2])
        image = tf.stack([s1, s2, s3], 2)
        return image

    @classmethod
    def sharpness(cls, image, factor):
        """Implements Sharpness function from PIL using TF ops."""
        orig_image = image
        image = tf.cast(image, tf.float32)
        # Make image 4D for conv operation.
        image = tf.expand_dims(image, 0)
        # SMOOTH PIL Kernel.
        kernel = tf.constant(
            [[1, 1, 1], [1, 5, 1], [1, 1, 1]], dtype=tf.float32,
            shape=[3, 3, 1, 1]) / 13.
        # Tile across channel dimension.
        kernel = tf.tile(kernel, [1, 1, 3, 1])
        strides = [1, 1, 1, 1]
        degenerate = tf.nn.depthwise_conv2d(
            image, kernel, strides, padding='VALID', dilations=[1, 1])
        degenerate = tf.clip_by_value(degenerate, 0.0, 255.0)
        degenerate = tf.squeeze(tf.cast(degenerate, tf.uint8), [0])

        # For the borders of the resulting image, fill in the values of the
        # original image.
        mask = tf.ones_like(degenerate)
        padded_mask = tf.pad(mask, [[1, 1], [1, 1], [0, 0]])
        padded_degenerate = tf.pad(degenerate, [[1, 1], [1, 1], [0, 0]])
        result = tf.where(tf.equal(padded_mask, 1), padded_degenerate, orig_image)

        # _blend the final result.
        return cls._blend(result, orig_image, factor)

    @classmethod
    def equalize(cls, image):
        """Implements Equalize function from PIL using TF ops."""
        def scale_channel(im, c):
            """Scale the data in the channel to implement equalize."""
            im = tf.cast(im[:, :, c], tf.int32)
            # Compute the histogram of the image channel.
            histo = tf.histogram_fixed_width(im, [0, 255], nbins=256)

            # For the purposes of computing the step, filter out the nonzeros.
            nonzero = tf.where(tf.not_equal(histo, 0))
            nonzero_histo = tf.reshape(tf.gather(histo, nonzero), [-1])
            step = (tf.reduce_sum(nonzero_histo) - nonzero_histo[-1]) // 255

            def build_lut(histo, step):
                # Compute the cumulative sum, shifting by step // 2
                # and then normalization by step.
                lut = (tf.cumsum(histo) + (step // 2)) // step
                # Shift lut, prepending with 0.
                lut = tf.concat([[0], lut[:-1]], 0)
                # Clip the counts to be in range.  This is done
                # in the C code for image.point.
                return tf.clip_by_value(lut, 0, 255)

            # If step is zero, return the original image.  Otherwise, build
            # lut from the full histogram and step and then index from it.
            result = tf.cond(tf.equal(step, 0),
                lambda: im,
                lambda: tf.gather(build_lut(histo, step), im))

            return tf.cast(result, tf.uint8)

        # Assumes RGB for now.  Scales each channel independently
        # and then stacks the result.
        s1 = scale_channel(image, 0)
        s2 = scale_channel(image, 1)
        s3 = scale_channel(image, 2)
        image = tf.stack([s1, s2, s3], 2)
        print(image)
        return image

    @classmethod
    def invert(cls, image):
        """Inverts the image pixels."""
        image = tf.convert_to_tensor(image)
        return 255 - image

    @classmethod
    def _randomly_negate_tensor(cls, tensor: tf.Tensor) -> tf.Tensor:
        """With 50% prob turn the tensor negative."""
        should_flip = tf.cast(tf.floor(tf.random.uniform([]) + 0.5), tf.bool)
        final_tensor = tf.cond(should_flip, lambda: tensor, lambda: -tensor)
        return final_tensor

    @classmethod
    def _rotate_level_to_arg(cls, level):
        level = (level/cls._MAX_LEVEL) * 30.
        level = cls._randomly_negate_tensor(level)
        return (float(level),)

    @classmethod
    def _shrink_level_to_arg(cls, level):
        """Converts level to ratio by which we shrink the image content."""
        if level == 0:
            return (1.0,)  # if level is zero, do not shrink the image
        # Maximum shrinking ratio is 2.9.
        level = 2. / (cls._MAX_LEVEL / level) + 0.9
        return (level,)

    @classmethod
    def _enhance_level_to_arg(cls, level):
        return ((level/cls._MAX_LEVEL) * 1.8 + 0.1,)

    @classmethod
    def _shear_level_to_arg(cls, level):
        level = (level/cls._MAX_LEVEL) * 0.3
        # Flip level to negative with 50% chance.
        level = cls._randomly_negate_tensor(level)
        return (level,)

    @classmethod
    def _translate_level_to_arg(cls, level, translate_const):
        level = (level/cls._MAX_LEVEL) * float(translate_const)
        # Flip level to negative with 50% chance.
        level = cls._randomly_negate_tensor(level)
        return (level,)

    @classmethod
    def level_to_arg(cls, cutout_const, translate_const):
        return {
            'AutoContrast': lambda level: (), #0
            'Equalize': lambda level: (),
            'Invert': lambda level: (),
            'Rotate': cls._rotate_level_to_arg,
            'Posterize': lambda level: (int((level/cls._MAX_LEVEL) * 4),),
            'Solarize': lambda level: (int((level/cls._MAX_LEVEL) * 256),), #5
            'SolarizeAdd': lambda level: (int((level/cls._MAX_LEVEL) * 110),),
            'Color': cls._enhance_level_to_arg,
            'Contrast': cls._enhance_level_to_arg,
            'Brightness': cls._enhance_level_to_arg,
            'Sharpness': cls._enhance_level_to_arg, #10
            'ShearX': cls._shear_level_to_arg,
            'ShearY': cls._shear_level_to_arg,
            'Cutout': lambda level: (int((level/cls._MAX_LEVEL) * cutout_const),),
            # pylint:disable=g-long-lambda
            'TranslateX': lambda level: cls._translate_level_to_arg(
                level, translate_const),
            'TranslateY': lambda level: cls._translate_level_to_arg(
                level, translate_const),
            # pylint:enable=g-long-lambda
        }


    @classmethod
    def randaugment_map(cls, num:int, magnitude:int, cutout_const:int=40, translate_const:int=100):
        """ RandAugmentを行うmap関数を返却する。

            RandAugment is from the paper https://arxiv.org/abs/1909.13719,
            
            Args:
                num (int): 何種類の拡張を実施するか。大抵は[1,3]の範囲が良い値。論文中ではNとして記載。
                magnitude: 拡張の大きさ。全ての拡張操作共通で使用される。論文中では(M)で記載。
                　　　　　　　大抵、[5,30]の範囲にベストが存在。
            Returns:
                [type]: [description]
        """

        available_ops = [
            'AutoContrast', 'Equalize', 'Invert', 'Rotate', 'Posterize',
            'Solarize', 'Color', 'Contrast', 'Brightness', 'Sharpness',
            'ShearX', 'ShearY', 'TranslateX', 'TranslateY', 'Cutout', 'SolarizeAdd']

        level = magnitude

        @tf.function
        def distort_each_ops(data):
            op_to_select = tf.random.uniform([], maxval=len(available_ops), dtype=tf.int32)
            # tf.print("ops ",op_to_select)
            ret = tf.switch_case(
                op_to_select,
                branch_fns={
                    0: lambda: cls.autocontrast(data["image"]),
                    1: lambda: cls.equalize(data["image"]),
                    2: lambda: cls.invert(data["image"]),
                    3: lambda: cls.rotate(data["image"], *(cls._rotate_level_to_arg(level))),
                    4: lambda: cls.posterize(data["image"] , int((level/cls._MAX_LEVEL)*4)),
                    5: lambda: cls.solarize(data["image"], int((level/cls._MAX_LEVEL)* 256)),
                    6: lambda: cls.solarize_add(data["image"],int((level/cls._MAX_LEVEL)* 110)),
                    7: lambda: cls.color(data["image"],*(cls._enhance_level_to_arg(level))),
                    8: lambda: cls.contrast(data["image"],*(cls._enhance_level_to_arg(level))),
                    9: lambda: cls.brightness(data["image"], *(cls._enhance_level_to_arg(level))),
                    10: lambda: cls.sharpness(data["image"], *(cls._enhance_level_to_arg(level))),
                    11: lambda: cls.shear_x(data["image"] , *(cls._shear_level_to_arg(level))),
                    12: lambda: cls.shear_y(data["image"] , *(cls._shear_level_to_arg(level))),
                    13: lambda: cls.translate_x(data["image"], *(cls._translate_level_to_arg(level, translate_const))),
                    14: lambda: cls.translate_y(data["image"], *(cls._translate_level_to_arg(level, translate_const))),
                    15: lambda: cls.cutout(data["image"], int((level/cls._MAX_LEVEL) * cutout_const)),
                },
                default= lambda: cls.autocontrast(data["image"], *())
            )
            return ret

        def distort_randaug(data):
            level = tf.cast(magnitude, tf.float32)
            # tf.print("level ", level)
            for i in range(num):
                image = distort_each_ops(data)
                data["image"] = image
            return data
        
        return distort_randaug




from abc import ABC, abstractclassmethod

import tensorflow as tf
import math
import os
import sys
from tftk import ResumeExecutor
from tftk import Context
from tftk import IS_SUSPEND_RESUME_TRAIN
from tftk import IS_ON_COLABOLATORY_WITH_GOOGLE_DRIVE
from tftk import Colaboratory

class CosineAnnealingScheduler(tf.keras.callbacks.Callback):
    """Cosine annealing scheduler.
    Cosine annealingを行います。このソースコードは以下を参照しています。
    https://github.com/4uiiurz1/keras-cosine-annealing
    
    """

    def __init__(self, T_max:int, eta_max:float, eta_min:float=0, verbose=0):
        """初期化をします。
        
        Arguments:
            T_max {int} -- 最大エポック数
            eta_max {float} -- 最大学習率
        
        Keyword Arguments:
            eta_min {float} -- 最小学習率 (default: {0})
            verbose {int} -- バーバスモード(default: {0})
        """
        super(CosineAnnealingScheduler, self).__init__()
        self.T_max = T_max
        self.eta_max = eta_max
        self.eta_min = eta_min
        self.verbose = verbose

    def on_epoch_begin(self, epoch:int, logs=None):
        """エポック開始前の処理を実施します。
        
        Arguments:
            epoch {int} -- エポック数
        
        Keyword Arguments:
            logs {[type]} -- [description] (default: {None})
        
        Raises:
            ValueError: [description]
        """
        if not hasattr(self.model.optimizer, 'lr'):
            raise ValueError('Optimizer must have a "lr" attribute.')
        lr = self.eta_min + (self.eta_max - self.eta_min) * (1 + math.cos(math.pi * epoch / self.T_max)) / 2
        tf.keras.backend.set_value(self.model.optimizer.lr, lr)
        if self.verbose > 0:
            print('\nEpoch %05d: CosineAnnealingScheduler setting learning '
                  'rate to %s.' % (epoch + 1, lr))

    def on_epoch_end(self, epoch, logs=None):
        """エポック終了時の処理を実施します。
        
        Arguments:
            epoch {int} -- エポック数
        
        Keyword Arguments:
            logs {[type]} -- [description] (default: {None})
        
        Raises:
            ValueError: [description]
        """
        logs = logs or {}
        logs['lr'] = tf.keras.backend.get_value(self.model.optimizer.lr)



class SuspendCallback(tf.keras.callbacks.Callback):
    """学習の経過を保存し、次回実行時に再開させる情報を記録する

    Args:
        tf ([type]): [description]
    """

    def __init__(self,  suspend_on_best:bool =True , monitor:str="val_loss",path:str=None):
        self.suspend_on_best = suspend_on_best
        self.monitor = monitor
        self.path = path
        if monitor == "val_loss" or monitor == "loss":
            self.best_value = sys.float_info.max
        else:
            self.best_value = sys.float_info.min
        
        super(SuspendCallback, self).__init__()


    def on_train_begin(self, epoch:int, logs=None):
        if IS_SUSPEND_RESUME_TRAIN() == True:
            exe = ResumeExecutor.get_instance()
            if exe.is_resumable_training():
                resume = exe.resume_values()
                e,l,b,_ = resume
                self.best_value = b
                self.model.optimizer.lr = l
                # print("Suspend Call Resume LR",l)

    def on_epoch_begin(self, epoch:int, logs=None):
        pass

    def set_model(self, model):
        self.model:tf.kearas.Model = model
            
    def on_epoch_end(self, epoch, logs=None):
        if logs == None:
            return
        
        value = logs[self.monitor]
        lr = logs["lr"]
        print("\n suspend on epoch end!!",epoch)
        
        exe = ResumeExecutor.get_instance()
        if self.monitor == "val_loss" or self.monitor == "loss":
            # print("\n monitor val_loss",self.best_value,",",value)
            if self.best_value > value:
                self.best_value = value
                exe.suspend(epoch,lr,self.best_value,self.model)
        else:
            if self.best_value < value:
                self.best_value = value
                exe.suspend(epoch,lr,self.best_value,self.model)
            
    def on_train_end(self, epoch:int, logs=None):
        exe = ResumeExecutor.get_instance()
        exe.training_completed()


class CallbackBuilder():
    @classmethod
    def get_callbacks(cls, tensorboard:bool=True, profile_batch:str=None, consine_annealing=False, cosine_init_lr=0.01, cosine_max_epochs = 60, reduce_lr_on_plateau=True,reduce_patience=5,reduce_factor=0.2,reduce_min=1e-6, early_stopping_patience=0):
        """よく利用するコールバックを設定します。
        
        Parameters:
            base_dir: ログを行うベースディレクトリ
            name: トレーニングの名前、base_dir/nameにデータを保存する
            resume: Resumeの有無



        Keyword Arguments:
            tensorboard_log_dir {str} -- tensorboardログを出力します。Noneの場合、出力しません。 (default: {None})
            profile_batch{str} -- プロファイルを行う際の開始バッチ、終了バッチを指定します。Noneの場合実行しません。
            save_weights {str} -- モデルを保存します。 (default: {"./tmp/model.hdf5"})
            monitor {str} -- [description] (default: {"val_acc"})
            max_epoch
        Returns:
            [type] -- [description]
        """

        context = Context.get_instance()
        base = context[Context.TRAINING_BASE_DIR]
        name = context[Context.TRAINING_NAME]

        traing_dir = base + os.path.sep + name

        # if name != None:
        #     base = base_dir + os.path.sep + name
        # else:
        #     files = os.listdir(base_dir)
        #     max_num = 0
        #     for f in files:
        #         try:
        #             tmp = int(f)
        #             if tmp +1 > max_num:
        #                 max_num = tmp + 1
        #         except Exception as e:
        #             pass
        #     base = base_dir + os.path.sep + str(max_num)
        
        if tf.io.gfile.exists(traing_dir)== False:
            tf.io.gfile.makedirs(traing_dir)

        callbacks = []

        if tensorboard is True:
            # print("Callback-TensorBoard")
            tensorboard_log_dir = traing_dir + os.path.sep + "log"
            if profile_batch != None:
                callbacks.append(tf.keras.callbacks.TensorBoard(log_dir=tensorboard_log_dir,profile_batch=profile_batch,histogram_freq=1))
            else:
                callbacks.append(tf.keras.callbacks.TensorBoard(log_dir=tensorboard_log_dir))

        # if save_weights == True:
        #     print("Callback-ModelCheckPoint")
        #     save_path = base + os.path.sep  + "model.hdf5"
        #     callbacks.append(tf.keras.callbacks.ModelCheckpoint(filepath=save_path,monitor="val_acc",save_best_only=True,save_weights_only=True))
        if consine_annealing == True:
            print("Callback-CosineAnnealing")
            cosine_annealer = CosineAnnealingScheduler(cosine_max_epochs,eta_max=cosine_init_lr,eta_min=0.0)
            callbacks.append(cosine_annealer)
        if reduce_lr_on_plateau ==True:
            print("Callback-ReduceOnPlateau")
            callbacks.append(tf.keras.callbacks.ReduceLROnPlateau(patience=reduce_patience,factor=reduce_factor,verbose=1,min_lr=reduce_min))
        if early_stopping_patience != 0:
            callbacks.append(tf.keras.callbacks.EarlyStopping(patience=early_stopping_patience,verbose=1))

        if IS_SUSPEND_RESUME_TRAIN() == True:
            print("Suspend Callback!!")
            callbacks.append(SuspendCallback())

        return callbacks




from typing import List
import tensorflow as tf

from tftk.image.dataset import ImageDatasetUtil
from tftk import IS_SUSPEND_RESUME_TRAIN, ResumeExecutor
from tftk import IS_ON_COLABOLATORY_WITH_GOOGLE_DRIVE, Colaboratory

class ImageTrain():

    def __init__(self):
        pass
    
    @classmethod
    def train_image_classification(
        cls,
        train_data:tf.data.Dataset, train_size:int, batch_size:int,
        validation_data:tf.data.Dataset, validation_size:int,
        shuffle_size:int,
        model:tf.keras.Model,
        callbacks:List[tf.keras.callbacks.Callback],
        optimizer:tf.keras.optimizers.Optimizer,
        loss:tf.keras.losses.Loss,
        max_epoch:int = 5, resume:bool = True):

        # dataset = dataset.shuffle(1024).batch(32).prefetch(tf.data.experimental.AUTOTUNE)

        train_data = train_data.map(ImageDatasetUtil.dict_to_classification_tuple(),num_parallel_calls=tf.data.experimental.AUTOTUNE).repeat()
        if shuffle_size != 0:
            train_data = train_data.shuffle(shuffle_size)
        train_data = train_data.batch(batch_size).prefetch(tf.data.experimental.AUTOTUNE)

        validation_data = validation_data.map(ImageDatasetUtil.dict_to_classification_tuple(),num_parallel_calls=tf.data.experimental.AUTOTUNE)
        validation_data = validation_data.repeat().batch(batch_size).prefetch(tf.data.experimental.AUTOTUNE)

        # model.compile(optimizer=optimizer,  loss=loss,  metrics="val_loss")

        model.compile(optimizer=optimizer, loss=loss, metrics=["acc"])

        model.summary()

        initial_epoch = 0
        exe = ResumeExecutor.get_instance()

        if IS_ON_COLABOLATORY_WITH_GOOGLE_DRIVE():
            print("colab training with google drive")
            Colaboratory.copy_resume_data_from_google_drive()
        else:
            print("google drive is not found.")

        if exe.is_resumable_training()==True:
            print("This is resume training!!")
            exe.resume_model(model)
            resume_val = exe.resume_values()
            initial_epoch, _, _,_  = resume_val
            initial_epoch = initial_epoch + 1
            print("resume epoch", initial_epoch, "max_epoch", max_epoch)
        else:
            if exe.is_train_ended()==True:
                print("Training is completed.")
                exit()
            else:
                print("Not resume training")

        """
        optimizer='rmsprop', loss=None, metrics=None, loss_weights=None,
        sample_weight_mode=None, weighted_metrics=None, target_tensors=None,
        distribute=None, **kwargs
        """
        steps_per_epoch = train_size//batch_size
        validation_steps = validation_size//batch_size
        model.fit(
            train_data,
            callbacks=callbacks,
            validation_data=validation_data,
            steps_per_epoch=steps_per_epoch,
            validation_steps=validation_steps,
            epochs=max_epoch, initial_epoch=initial_epoch)

"""
fit(
    x=None, y=None, batch_size=None, epochs=1, verbose=1, callbacks=None,
    validation_split=0.0, validation_data=None, shuffle=True, class_weight=None,
    sample_weight=None, initial_epoch=0, steps_per_epoch=None,
    validation_steps=None, validation_freq=1, max_queue_size=10, workers=1,
    use_multiprocessing=False, **kwargs
)
"""


from . classification import AbstractClassificationModel
from . classification import KerasResNet50V2
from . classification import KerasResNet50
from . classification import KerasMobileNetV2

from . efficientnet import KerasEfficientNetB0
from . efficientnet import KerasEfficientNetB1
from . efficientnet import KerasEfficientNetB2
from . efficientnet import KerasEfficientNetB3
from . efficientnet import KerasEfficientNetB4
from . efficientnet import KerasEfficientNetB5
from . efficientnet import KerasEfficientNetB6
from . efficientnet import KerasEfficientNetB7
from . efficientnet import EfficientNetB0

# from . efficientnet import KerasEfficientNetB1
# from . efficientnet import KerasEfficientNetB2

# from . small_resnet_keras_contrib import ResNet18
# from . small_resnet_keras_contrib import ResNet34
from . resnet import ResNet18
from . resnet import ResNet34
from . resnet import ResNet50

__all__ = [
    'AbstractClassificationModel'
    'KerasResNet50',
    'KerasResNet50V2',
    'KerasMobileNetV2',
    'KerasEfficientNetB0',
    'KerasEfficientNetB1',
    'KerasEfficientNetB2',
    'KerasEfficientNetB3',
    'KerasEfficientNetB4',
    'KerasEfficientNetB5',
    'KerasEfficientNetB6',
    'KerasEfficientNetB7',
    'ResNet18',
    'ResNet34',
    'ResNet50',
    'EfficientNetB0'
]

"""
    'ConvAutoEncoder',
    'BaseModel',
    'SimpleBaseModel',
    'ResNet18',
    'ResNet34'
    'ResNet50V2',

"""


import os
import tensorflow as tf
from . context import Context
from . colab import IS_ON_COLABOLATORY_WITH_GOOGLE_DRIVE, Colaboratory

def ENABLE_SUSPEND_RESUME_TRAIN():
    context = Context.get_instance()
    context[Context.SUSPEND_RESUME] = True

def IS_SUSPEND_RESUME_TRAIN():
    context = Context.get_instance()
    return context[Context.SUSPEND_RESUME]

class ResumeExecutor():

    instance = None

    RESUME_FILE = 'RESUME_INFO.txt'

    MODEL_FILE = 'model.h5'

    def __init__(self):
        context = Context.get_instance()
        self.base_dir = context[Context.TRAINING_BASE_DIR]
        self.name = context[Context.TRAINING_NAME]
        self.path = self.base_dir + os.path.sep + self.name

    @classmethod
    def get_instance(cls)->'ResumeExecutor':
        if cls.instance == None:
            cls.instance = ResumeExecutor()
        
        return cls.instance
    
    def _load_values(self)->(int,float,bool):
        f = tf.io.gfile.GFile(self._get_resume_path(),mode="r")
        s = f.read()
        splited = s.split(",")
        epoch = int(splited[0])
        lr = float(splited[1])
        best = float(splited[2])
        end = (splited[3]=="True")
        f.close()
        ret = (epoch,lr,best,end)
        # print(ret)
        return ret

    def is_train_ended(self)->bool:
        if tf.io.gfile.exists(self.path) == True:
            _, _, _, end =self._load_values()
            return end
        else:
            return False
    
    def is_resumable_training(self)->bool:
        if tf.io.gfile.exists(self._get_model_path()):
            return not self.is_train_ended()
        else:
            return False

    def _check_dir(self):
        if tf.io.gfile.exists(self.path) == False:
            tf.io.gfile.makedirs(self.path)

    def _get_model_path(self)->str:
        return self.path + os.path.sep + ResumeExecutor.MODEL_FILE
    
    def _get_resume_path(self)->str:
        return self.path + os.path.sep + ResumeExecutor.RESUME_FILE

    def resume_values(self)->(int,float):
        lv = self._load_values()
        return lv

    def resume_model(self,model:tf.keras.Model):
        model.load_weights(self._get_model_path())

    def suspend(self,epoch,lr,best,model):
        print("\nSusupend!!!")
        self._check_dir()
        file = self._get_resume_path()
        f = tf.io.gfile.GFile(file, mode="w")
        c = str(epoch)+","+str(lr)+"," + str(best) + "," +str(False)
        print("suspend ",file,c)
        resume_info = f.write(c)
        f.flush()
        f.close()
        if model != None:
            model.save_weights(self._get_model_path())
        
        if IS_ON_COLABOLATORY_WITH_GOOGLE_DRIVE():
            Colaboratory.copy_suspend_data_from_colab()

    def training_completed(self):
        self._check_dir()
        f = tf.io.gfile.GFile(self._get_resume_path(), mode="w")
        resume_info = f.write(str(0)+","+str(0.0)+"," +str(0.0)+","+ str(True))
        f.close()
        
        if IS_ON_COLABOLATORY_WITH_GOOGLE_DRIVE():
            Colaboratory.copy_suspend_data_from_colab()






from . base import *
from . classification import *
from . anomaly_detection import *
from . utility import *

all = [
    'BaseDataset',
    'MVTecAd',
    'Mnist',
    'Cifar10',
    'Food101',
    'PatchCamelyon',
    'ImageLabelFolderDataset',
    'ImageNetResized',
    'ImageDatasetUtil',
    'ImageNet2012',
    'AbstractClassificationModel',
    'ImageCrawler',
    'Place365Small'
]


from abc import ABC, abstractclassmethod
import tensorflow as tf

class BaseDataset(ABC):
    """Datasetのベースクラス

    """

    @abstractclassmethod
    def get_train_dataset(cls, **kwargs)->(tf.data.Dataset, int):
        """学習用のデータセットを取得する
        
        Arguments:
            cls -- クラスオブジェクト
        
        Returns:
            データセットとデータセットのサイズを持ったタプル
        """
        raise NotImplementedError()

    @abstractclassmethod
    def get_validation_dataset(cls, **kwargs)->(tf.data.Dataset, int):
        """Validation用のデータセットを取得する
        
        Arguments:
            cls -- クラスオブジェクト
        
        Returns:
            データセットとデータセットのサイズを持ったタプル
        """
        raise NotImplementedError()


    @abstractclassmethod
    def get_test_dataset(cls, **kwargs)->(tf.data.Dataset, int):
        """テスト用のデータセットを取得する
        
        Arguments:
            cls -- クラスオブジェクト
        
        Returns:
            データセットとデータセットのサイズを持ったタプル
        """
        raise NotImplementedError()





import tensorflow as tf
from typing import Dict,List, Callable

class ImageDatasetUtil():

    def __init__(self):
        pass
    
    @classmethod
    def devide_train_validation(cls, dataset:tf.data.Dataset, length:int, ratio:float)->((tf.data.Dataset,int),(tf.data.Dataset,int)):
        """学習用と検証用のデータセットに分割する。
        
        Arguments:
            dataset -- 分割するデータセット
            length -- データセットの長さ
            ratio -- 学習用データセットの割合
        
        Returns:
            (Dataset,int),(Dataset,int) -- (学習用のデータセット、データセットのサイズ),(検証用データセット,データセットのサイズ)
        """

        train_size = int(length * ratio)
        validation_size = length - train_size
        train_set = dataset.take(train_size).skip(validation_size)
        validation_set = dataset.skip(train_size).take(validation_size)
        return ((train_set,train_size),(validation_set,validation_size))


    @classmethod
    def k_fold_cross_validation_dataset(cls, dataset:tf.data.Dataset, length:int, ratio:float)->((tf.data.Dataset,int),(tf.data.Dataset,int)):
        pass        

    @classmethod
    def count_image_dataset(cls, dataset:tf.data.Dataset)->(int, Dict):
        """データセットのラベルごとの数を数える。
        
        Arguments:
            cls {[type]} -- [description]
            Dict {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        total = 0
        count = {}
        for d in dataset:
            y = d["label"].numpy()
            c = count.get(y,0) + 1
            count[y] = c
            total = total + 1
        print(total)
        print(count)
        return total, count
    
    @classmethod
    def one_hot(cls, classes:int):
        """データセットに指定クラス数のone-hot処理を与えるmap関数

        
        Arguments:
            classes {int} -- [description]
        
        Returns:
            [type] -- [description]
        """
        def one_hot_map(data):
            data["label"] = tf.one_hot(data["label"], classes)
            return data
        return one_hot_map

    @classmethod
    def image_reguralization(cls, offset:float=0.0):
        """画像データを正則化するmap関数を作成する。
        
        Keyword Arguments:
            offset {float} -- [description] (default: {0.0})
        
        Returns:
            [type] -- [description]
        """
        def image_reguralization_map(data):
            data["image"]=tf.cast(data["image"], tf.float32)
            data["image"]=data["image"]/255.0 - offset
            return data
        return image_reguralization_map
    
    @classmethod
    def dataset_init_classification(cls, classes, offset:float=0.0)->Callable[[Dict],Dict]:
        """ 画像を数値/255.0、ラベルのont-hot化、オフセットの除去
        
        Arguments:
            classes {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        def __dataset_init_classification(data:Dict)->Dict:
            data["label"] = tf.one_hot(data["label"], classes)
            data["image"] = tf.cast(data["image"], tf.float32)
            data["image"] = (data["image"]/255.0) - offset
            return data
        return __dataset_init_classification
    
    @classmethod
    def dict_to_classification_tuple(cls):
        """dict形式のデータをtf.keras.Model.fit()用に変換する
        
        Returns:
            [type]: [description]
        """
        def __dict_to_classification_tuple(data:Dict)->(tf.Tensor,tf.Tensor):
            return (data["image"], data["label"])
        return __dict_to_classification_tuple
    
    @classmethod
    def dict_to_autoencoder_tuple(cls):
        """dict形式のデータをtf.keras.Model.fit()用に変換する
        
        Returns:
            [type]: [description]
        """
        pass
        


    @classmethod
    def resize_with_crop_or_pad(cls, hight:int, width:int)->Callable[[Dict],Dict]:
        """データセットの画像をクロップないしはパッドしてリサイズする。
        
        Args:
            h (int): リサイズする画像の高さ
            w (int): リサイズ後の画像の幅
        
        Returns:
            Mapする関数: Dataset.map()に適用する関数
        """
        def resize_with_crop_or_pad_map(data:Dict)->Dict:
            data["image"] = tf.image.resize_with_crop_or_pad(data["image"], hight, width)
            return data
        return resize_with_crop_or_pad_map
    
    @classmethod
    def resize(cls, h:int, w:int)->Callable[[Dict],Dict]:
        """データセットの画像をリサイズする。
        
        Args:
            h (int): リサイズする画像の高さ
            w (int): リサイズ後の画像の幅
        
        Returns:
            Mapする関数: Dataset.map()に適用する関数
        """

        def __resize(data:Dict)->Dict:
            data["image"] = tf.image.resize(data["image"], (h,w))
            return data
        return __resize


import os
from os.path import expanduser

import tensorflow as tf
import tensorflow_datasets as tfds


class MVTecAd():
    """MVTec Adをデータセット

    ファイルを ~/dataset/mvtec_ad/(bottole etc..)に解凍します。
    
    """

    def __init__(self):
        """ 初期化
        """
        pass

    @classmethod
    def get_train_dataset(cls, type="bottle")->(tf.data.Dataset, int):
        manual_dir = expanduser("~") + os.path.sep + "dataset" + os.path.sep + "mvtec_ad"
        print(manual_dir)
        builder = tfds.image.ImageLabelFolder(type)

        dl_config = tfds.download.DownloadConfig(manual_dir=manual_dir)
        builder.download_and_prepare(download_config=dl_config)
        ds = builder.as_dataset(split='train', shuffle_files=False)
        len = builder.info.splits['train'].num_examples  # Splits, num examples,... automatically extracted
        return ds, len

    @classmethod
    def get_test_dataset(cls, type="bottle")->(tf.data.Dataset,int):
        """test用データセットを取得する。
        
        Arguments:
            int {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        dir = expanduser("~") + os.path.sep + "dataset" + os.path.sep + "mvtec_ad" + os.path.sep + type + os.path.sep + "train" + os.path.sep + "good"
        print(dir)
        builder = tfds.image.ImageLabelFolder(dir)

        dl_config = tfds.download.DownloadConfig()
        builder.download_and_prepare(download_config=dl_config)
        print(builder.info)  # Splits, num examples,... automatically extracted
        ds, info = builder.as_dataset(split='test', shuffle_files=True, with_info=True)
        len = builder.info.splits['test'].num_examples
        return ds, len
    



from . activation import USE_MISH_AS_RELU
from . mixed_precision import USE_MIXED_PRECISION
from . mixed_precision import IS_MIXED_PRECISION
from . resume import ENABLE_SUSPEND_RESUME_TRAIN
from . resume import IS_SUSPEND_RESUME_TRAIN
from . resume import ResumeExecutor
from . context import Context
from . colab import Colaboratory, IS_ON_COLABOLATORY_WITH_GOOGLE_DRIVE

__all__ = [
    'USE_MIXED_PRECISION', 
    'IS_MIXED_PRECISION',
    'USE_MISH_AS_RELU',
    'ENABLE_SUSPEND_RESUME_TRAIN',
    'IS_SUSPEND_RESUME_TRAIN',
    'ResumeExecutor',
    'Context'
]


import tensorflow as tf

from . context import Context

mixed_precision = False


def USE_MIXED_PRECISION():
    Context.get_instance()[Context.MIXED_PRECISION]=True
    policy = tf.keras.mixed_precision.experimental.Policy('mixed_float16')
    tf.keras.mixed_precision.experimental.set_policy(policy)

def IS_MIXED_PRECISION()->bool:
    policy = tf.keras.mixed_precision.experimental.global_policy()
    context = Context.get_instance()
    return context[Context.MIXED_PRECISION]

"""
--- ポリシーの設定
policy = mixed_precision.Policy('mixed_float16')
mixed_precision.set_policy(policy)
---

--optimizerをdynamicでラップ
optimizer = keras.optimizers.RMSprop()
optimizer = mixed_precision.LossScaleOptimizer(optimizer, loss_scale='dynamic')
--
"""



from typing import Dict

class Context(Dict):

    instance = None

    MIXED_PRECISION = "MIXED_PRECISION"
    SUSPEND_RESUME = "SUSPEND_RESUME"

    TRAINING_BASE_DIR = "TRAINING_BASE_DIR"
    TRAINING_NAME = "TRAINING_NAME"

    def __init__(self):
        super(Context, self).__init__()
    
    @classmethod
    def init_context(cls, **kwargs):
        ctx = Context.get_instance()
        for k in kwargs.keys():
            ctx[k] = kwargs[k]
    
    @classmethod
    def get_instance(cls)->'Context':
        if cls.instance == None:
            cls.instance = Context()
            cls.instance[Context.TRAINING_BASE_DIR] ="tmp"
            cls.instance[Context.TRAINING_NAME] = "temp_name"
            cls.instance[Context.MIXED_PRECISION] = False
            cls.instance[Context.SUSPEND_RESUME] = False
        return cls.instance
   



import tensorflow as tf

from abc import ABC, abstractclassmethod, abstractmethod
import tensorflow as tf


class AbstractClassificationModel(ABC):

    def __init__(self):
        pass

    @abstractclassmethod
    def get_base_model(cls,input:(int,int,int),include_top:bool,weights:str,**kwargs)->tf.keras.Model:
        raise NotImplementedError()

    @classmethod
    def get_model(cls, input_shape:(int,int,int), include_top=False, weights=None, classes=1000, **kwargs)-> tf.keras.Model:
        base = cls.get_base_model(input_shape,include_top, weights, **kwargs)
        model = ClassificationModel.get_classify_model(base,classes)
        return model

class SimpleClassificationModel(AbstractClassificationModel):
    def __init__(self):
        pass
    
    @classmethod
    def get_base_model(cls,input_shape:(int,int,int),include_top:bool,weights:str,**kwargs) -> tf.keras.Model:
        inputs = tf.keras.Input(shape=input_shape,name='image')
        x = tf.keras.layers.Conv2D(filters=32,kernel_size=(3,3),padding="same")(inputs)
        x = tf.keras.layers.Activation('relu')(x)
        x = tf.keras.layers.MaxPool2D((2,2))(x)

        x = tf.keras.layers.Conv2D(filters=64,kernel_size=(3,3),padding="same")(x) 
        x = tf.keras.layers.Activation('relu')(x)
        x = tf.keras.layers.MaxPool2D((2,2))(x)

        x = tf.keras.layers.Conv2D(filters=128,kernel_size=(3,3),padding='same')(x) 
        x = tf.keras.layers.Activation('relu')(x)
        x = tf.keras.layers.MaxPool2D((2,2))(x)

        x = tf.keras.layers.Conv2D(filters=128,kernel_size=(3,3),padding="same")(x) 
        x = tf.keras.layers.Activation('relu')(x)
        x = tf.keras.layers.MaxPool2D((2,2))(x)
        model = tf.keras.Model(inputs=inputs,outputs=x)
        return model


class KerasResNet50(AbstractClassificationModel):
    def __init__(self):
        pass
    
    @classmethod
    def get_base_model(cls,input_shape:(int,int,int),include_top:bool,weights:str,**kwargs) -> tf.keras.Model:
        model = tf.keras.applications.ResNet50(include_top=False, weights=None, input_shape=input_shape,pooling="avg")
        return model

class KerasResNet50V2(AbstractClassificationModel):

    def __init__(self):
        pass
    
    @classmethod
    def get_base_model(cls,input_shape:(int,int,int),include_top:bool,weights:str,**kwargs) -> tf.keras.Model:
        model = tf.keras.applications.ResNet50V2(include_top=False, weights=None, input_shape=input_shape,pooling="avg")
        return model

class KerasMobileNetV2(AbstractClassificationModel):

    def __init__(self):
        pass
    
    @classmethod
    def get_base_model(cls,input_shape:(int,int,int),include_top:bool,weights:str,**kwargs) -> tf.keras.Model:
        model = tf.keras.applications.MobileNetV2(include_top=False, weights=None, input_shape=input_shape,pooling="avg")
        return model

class KerasEfficentNetB0(AbstractClassificationModel):

    def __init__(self):
        pass
    
    @classmethod
    def get_base_model(cls,input_shape:(int,int,int),include_top:bool,weights:str,**kwargs) -> tf.keras.Model:
        model = tf.keras.applications.EfficientNetB0(include_top=False, weights=None, input_shape=input_shape, pooling="avg")
        return model


class ClassificationModel():
    """単純なSoftmaxによる分類モデルクラス

    """
    def __init__(self):
        pass

    @classmethod
    def get_classify_model(cls, base_model:tf.keras.Model, classes:int)->tf.keras.Model:        
        input = base_model.input
        last = base_model.output


        x = tf.keras.layers.GlobalAveragePooling2D()(last)
        
        # x = tf.keras.layers.Flatten(name='classify-1')(last)
        # x = tf.keras.layers.Dense(512, kernel_initializer='he_normal')(x)
        # x = tf.keras.layers.Activation('relu')(x)

        if classes != 2:
            x = tf.keras.layers.Dense(classes,dtype='float32',kernel_initializer='he_normal')(x)
            x = tf.keras.layers.Activation('softmax',dtype='float32')(x)
        else:
            x = tf.keras.layers.Dense(classes,dtype='float32',kernel_initializer='he_normal')(x)
            x = tf.keras.layers.Activation('sigmoid',dtype='float32')(x)
        return tf.keras.Model(input,x)




# EfficientNet is not is not export in TF2.2.0
# This is taken from keras applications.
# https://github.com/keras-team/keras-applications/blob/master/keras_applications/efficientnet.py

from . classification import AbstractClassificationModel
from tftk.activation import Mish

"""
COPYRIGHT

Copyright (c) 2016 - 2018, the respective contributors.
All rights reserved.

Each contributor holds copyright over their respective contributions.
The project versioning (Git) records all such contribution source information.
The initial code of this repository came from https://github.com/keras-team/keras
(the Keras repository), hence, for author information regarding commits
that occured earlier than the first commit in the present repository,
please see the original Keras repository.

LICENSE

The MIT License (MIT)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

"""EfficientNet models for Keras.
# Reference paper
- [EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks]
  (https://arxiv.org/abs/1905.11946) (ICML 2019)
# Reference implementation
- [TensorFlow]
  (https://github.com/tensorflow/tpu/tree/master/models/official/efficientnet)
"""

# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function

import os
import math

import tensorflow as tf


# from . import correct_pad
# from . import get_submodules_from_kwargs
# from . import imagenet_utils
# from .imagenet_utils import decode_predictions
# from .imagenet_utils import _obtain_input_shape


backend = None
layers = None
models = None
keras_utils = None

#     backend, layers, models, keras_utils = get_submodules_from_kwargs(kwargs)
def get_submodules_from_kwargs(kwargs):
    return (tf.keras.backend, tf.keras.layers, tf.keras.models, tf.keras.utils)

def _obtain_input_shape(input_shape,
                        default_size,
                        min_size,
                        data_format,
                        require_flatten,
                        weights=None):
    """Internal utility to compute/validate a model's input shape.
    # Arguments
        input_shape: Either None (will return the default network input shape),
            or a user-provided shape to be validated.
        default_size: Default input width/height for the model.
        min_size: Minimum input width/height accepted by the model.
        data_format: Image data format to use.
        require_flatten: Whether the model is expected to
            be linked to a classifier via a Flatten layer.
        weights: One of `None` (random initialization)
            or 'imagenet' (pre-training on ImageNet).
            If weights='imagenet' input channels must be equal to 3.
    # Returns
        An integer shape tuple (may include None entries).
    # Raises
        ValueError: In case of invalid argument values.
    """
    if weights != 'imagenet' and input_shape and len(input_shape) == 3:
        if data_format == 'channels_first':
            if input_shape[0] not in {1, 3}:
                warnings.warn(
                    'This model usually expects 1 or 3 input channels. '
                    'However, it was passed an input_shape with ' +
                    str(input_shape[0]) + ' input channels.')
            default_shape = (input_shape[0], default_size, default_size)
        else:
            if input_shape[-1] not in {1, 3}:
                warnings.warn(
                    'This model usually expects 1 or 3 input channels. '
                    'However, it was passed an input_shape with ' +
                    str(input_shape[-1]) + ' input channels.')
            default_shape = (default_size, default_size, input_shape[-1])
    else:
        if data_format == 'channels_first':
            default_shape = (3, default_size, default_size)
        else:
            default_shape = (default_size, default_size, 3)
    if weights == 'imagenet' and require_flatten:
        if input_shape is not None:
            if input_shape != default_shape:
                raise ValueError('When setting `include_top=True` '
                                 'and loading `imagenet` weights, '
                                 '`input_shape` should be ' +
                                 str(default_shape) + '.')
        return default_shape
    if input_shape:
        if data_format == 'channels_first':
            if input_shape is not None:
                if len(input_shape) != 3:
                    raise ValueError(
                        '`input_shape` must be a tuple of three integers.')
                if input_shape[0] != 3 and weights == 'imagenet':
                    raise ValueError('The input must have 3 channels; got '
                                     '`input_shape=' + str(input_shape) + '`')
                if ((input_shape[1] is not None and input_shape[1] < min_size) or
                   (input_shape[2] is not None and input_shape[2] < min_size)):
                    raise ValueError('Input size must be at least ' +
                                     str(min_size) + 'x' + str(min_size) +
                                     '; got `input_shape=' +
                                     str(input_shape) + '`')
        else:
            if input_shape is not None:
                if len(input_shape) != 3:
                    raise ValueError(
                        '`input_shape` must be a tuple of three integers.')
                if input_shape[-1] != 3 and weights == 'imagenet':
                    raise ValueError('The input must have 3 channels; got '
                                     '`input_shape=' + str(input_shape) + '`')
                if ((input_shape[0] is not None and input_shape[0] < min_size) or
                   (input_shape[1] is not None and input_shape[1] < min_size)):
                    raise ValueError('Input size must be at least ' +
                                     str(min_size) + 'x' + str(min_size) +
                                     '; got `input_shape=' +
                                     str(input_shape) + '`')
    else:
        if require_flatten:
            input_shape = default_shape
        else:
            if data_format == 'channels_first':
                input_shape = (3, None, None)
            else:
                input_shape = (None, None, 3)
    if require_flatten:
        if None in input_shape:
            raise ValueError('If `include_top` is True, '
                             'you should specify a static `input_shape`. '
                             'Got `input_shape=' + str(input_shape) + '`')
    return input_shape



def correct_pad(backend, inputs, kernel_size):
    """Returns a tuple for zero-padding for 2D convolution with downsampling.
    # Arguments
        input_size: An integer or tuple/list of 2 integers.
        kernel_size: An integer or tuple/list of 2 integers.
    # Returns
        A tuple.
    """
    img_dim = 2 if backend.image_data_format() == 'channels_first' else 1
    input_size = backend.int_shape(inputs)[img_dim:(img_dim + 2)]

    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size)

    if input_size[0] is None:
        adjust = (1, 1)
    else:
        adjust = (1 - input_size[0] % 2, 1 - input_size[1] % 2)

    correct = (kernel_size[0] // 2, kernel_size[1] // 2)

    return ((correct[0] - adjust[0], correct[0]),
            (correct[1] - adjust[1], correct[1]))



BASE_WEIGHTS_PATH = (
    'https://github.com/Callidior/keras-applications/'
    'releases/download/efficientnet/')
WEIGHTS_HASHES = {
    'b0': ('e9e877068bd0af75e0a36691e03c072c',
           '345255ed8048c2f22c793070a9c1a130'),
    'b1': ('8f83b9aecab222a9a2480219843049a1',
           'b20160ab7b79b7a92897fcb33d52cc61'),
    'b2': ('b6185fdcd190285d516936c09dceeaa4',
           'c6e46333e8cddfa702f4d8b8b6340d70'),
    'b3': ('b2db0f8aac7c553657abb2cb46dcbfbb',
           'e0cf8654fad9d3625190e30d70d0c17d'),
    'b4': ('ab314d28135fe552e2f9312b31da6926',
           'b46702e4754d2022d62897e0618edc7b'),
    'b5': ('8d60b903aff50b09c6acf8eaba098e09',
           '0a839ac36e46552a881f2975aaab442f'),
    'b6': ('a967457886eac4f5ab44139bdd827920',
           '375a35c17ef70d46f9c664b03b4437f2'),
    'b7': ('e964fd6e26e9a4c144bcb811f2a10f20',
           'd55674cc46b805f4382d18bc08ed43c1')
}


DEFAULT_BLOCKS_ARGS = [
    {'kernel_size': 3, 'repeats': 1, 'filters_in': 32, 'filters_out': 16,
     'expand_ratio': 1, 'id_skip': True, 'strides': 1, 'se_ratio': 0.25},
    {'kernel_size': 3, 'repeats': 2, 'filters_in': 16, 'filters_out': 24,
     'expand_ratio': 6, 'id_skip': True, 'strides': 2, 'se_ratio': 0.25},
    {'kernel_size': 5, 'repeats': 2, 'filters_in': 24, 'filters_out': 40,
     'expand_ratio': 6, 'id_skip': True, 'strides': 2, 'se_ratio': 0.25},
    {'kernel_size': 3, 'repeats': 3, 'filters_in': 40, 'filters_out': 80,
     'expand_ratio': 6, 'id_skip': True, 'strides': 2, 'se_ratio': 0.25},
    {'kernel_size': 5, 'repeats': 3, 'filters_in': 80, 'filters_out': 112,
     'expand_ratio': 6, 'id_skip': True, 'strides': 1, 'se_ratio': 0.25},
    {'kernel_size': 5, 'repeats': 4, 'filters_in': 112, 'filters_out': 192,
     'expand_ratio': 6, 'id_skip': True, 'strides': 2, 'se_ratio': 0.25},
    {'kernel_size': 3, 'repeats': 1, 'filters_in': 192, 'filters_out': 320,
     'expand_ratio': 6, 'id_skip': True, 'strides': 1, 'se_ratio': 0.25}
]

CONV_KERNEL_INITIALIZER = {
    'class_name': 'VarianceScaling',
    'config': {
        'scale': 2.0,
        'mode': 'fan_out',
        # EfficientNet actually uses an untruncated normal distribution for
        # initializing conv layers, but keras.initializers.VarianceScaling use
        # a truncated distribution.
        # We decided against a custom initializer for better serializability.
        'distribution': 'normal'
    }
}

DENSE_KERNEL_INITIALIZER = {
    'class_name': 'VarianceScaling',
    'config': {
        'scale': 1. / 3.,
        'mode': 'fan_out',
        'distribution': 'uniform'
    }
}


def swish(x):
    """Swish activation function.
    # Arguments
        x: Input tensor.
    # Returns
        The Swish activation: `x * sigmoid(x)`.
    # References
        [Searching for Activation Functions](https://arxiv.org/abs/1710.05941)
    """
    if backend.backend() == 'tensorflow':
        try:
            # The native TF implementation has a more
            # memory-efficient gradient implementation
            return backend.tf.nn.swish(x)
        except AttributeError:
            pass

    return x * backend.sigmoid(x)


def block(inputs, activation_fn=swish, drop_rate=0., name='',
          filters_in=32, filters_out=16, kernel_size=3, strides=1,
          expand_ratio=1, se_ratio=0., id_skip=True):
    """A mobile inverted residual block.
    # Arguments
        inputs: input tensor.
        activation_fn: activation function.
        drop_rate: float between 0 and 1, fraction of the input units to drop.
        name: string, block label.
        filters_in: integer, the number of input filters.
        filters_out: integer, the number of output filters.
        kernel_size: integer, the dimension of the convolution window.
        strides: integer, the stride of the convolution.
        expand_ratio: integer, scaling coefficient for the input filters.
        se_ratio: float between 0 and 1, fraction to squeeze the input filters.
        id_skip: boolean.
    # Returns
        output tensor for the block.
    """
    bn_axis = 3 if backend.image_data_format() == 'channels_last' else 1

    # Expansion phase
    filters = filters_in * expand_ratio
    if expand_ratio != 1:
        x = layers.Conv2D(filters, 1,
                          padding='same',
                          use_bias=False,
                          kernel_initializer=CONV_KERNEL_INITIALIZER,
                          name=name + 'expand_conv')(inputs)
        x = layers.BatchNormalization(axis=bn_axis, name=name + 'expand_bn')(x)
        x = layers.Activation(activation_fn, name=name + 'expand_activation')(x)
    else:
        x = inputs

    # Depthwise Convolution
    if strides == 2:
        x = layers.ZeroPadding2D(padding=correct_pad(backend, x, kernel_size),
                                 name=name + 'dwconv_pad')(x)
        conv_pad = 'valid'
    else:
        conv_pad = 'same'
    x = layers.DepthwiseConv2D(kernel_size,
                               strides=strides,
                               padding=conv_pad,
                               use_bias=False,
                               depthwise_initializer=CONV_KERNEL_INITIALIZER,
                               name=name + 'dwconv')(x)
    x = layers.BatchNormalization(axis=bn_axis, name=name + 'bn')(x)
    x = layers.Activation(activation_fn, name=name + 'activation')(x)

    # Squeeze and Excitation phase
    if 0 < se_ratio <= 1:
        filters_se = max(1, int(filters_in * se_ratio))
        se = layers.GlobalAveragePooling2D(name=name + 'se_squeeze')(x)
        if bn_axis == 1:
            se = layers.Reshape((filters, 1, 1), name=name + 'se_reshape')(se)
        else:
            se = layers.Reshape((1, 1, filters), name=name + 'se_reshape')(se)
        se = layers.Conv2D(filters_se, 1,
                           padding='same',
                           activation=activation_fn,
                           kernel_initializer=CONV_KERNEL_INITIALIZER,
                           name=name + 'se_reduce')(se)
        se = layers.Conv2D(filters, 1,
                           padding='same',
                           activation='sigmoid',
                           kernel_initializer=CONV_KERNEL_INITIALIZER,
                           name=name + 'se_expand')(se)
        if backend.backend() == 'theano':
            # For the Theano backend, we have to explicitly make
            # the excitation weights broadcastable.
            se = layers.Lambda(
                lambda x: backend.pattern_broadcast(x, [True, True, True, False]),
                output_shape=lambda input_shape: input_shape,
                name=name + 'se_broadcast')(se)
        x = layers.multiply([x, se], name=name + 'se_excite')

    # Output phase
    x = layers.Conv2D(filters_out, 1,
                      padding='same',
                      use_bias=False,
                      kernel_initializer=CONV_KERNEL_INITIALIZER,
                      name=name + 'project_conv')(x)
    x = layers.BatchNormalization(axis=bn_axis, name=name + 'project_bn')(x)
    if (id_skip is True and strides == 1 and filters_in == filters_out):
        if drop_rate > 0:
            x = layers.Dropout(drop_rate,
                               noise_shape=(None, 1, 1, 1),
                               name=name + 'drop')(x)
        x = layers.add([x, inputs], name=name + 'add')

    return x


def EfficientNet(width_coefficient,
                 depth_coefficient,
                 default_size,
                 dropout_rate=0.2,
                 drop_connect_rate=0.2,
                 depth_divisor=8,
                 activation_fn=swish,
                 blocks_args=DEFAULT_BLOCKS_ARGS,
                 model_name='efficientnet',
                 include_top=True,
                 weights='imagenet',
                 input_tensor=None,
                 input_shape=None,
                 pooling=None,
                 classes=1000,
                 **kwargs):
    """Instantiates the EfficientNet architecture using given scaling coefficients.
    Optionally loads weights pre-trained on ImageNet.
    Note that the data format convention used by the model is
    the one specified in your Keras config at `~/.keras/keras.json`.
    # Arguments
        width_coefficient: float, scaling coefficient for network width.
        depth_coefficient: float, scaling coefficient for network depth.
        default_size: integer, default input image size.
        dropout_rate: float, dropout rate before final classifier layer.
        drop_connect_rate: float, dropout rate at skip connections.
        depth_divisor: integer, a unit of network width.
        activation_fn: activation function.
        blocks_args: list of dicts, parameters to construct block modules.
        model_name: string, model name.
        include_top: whether to include the fully-connected
            layer at the top of the network.
        weights: one of `None` (random initialization),
              'imagenet' (pre-training on ImageNet),
              or the path to the weights file to be loaded.
        input_tensor: optional Keras tensor
            (i.e. output of `layers.Input()`)
            to use as image input for the model.
        input_shape: optional shape tuple, only to be specified
            if `include_top` is False.
            It should have exactly 3 inputs channels.
        pooling: optional pooling mode for feature extraction
            when `include_top` is `False`.
            - `None` means that the output of the model will be
                the 4D tensor output of the
                last convolutional layer.
            - `avg` means that global average pooling
                will be applied to the output of the
                last convolutional layer, and thus
                the output of the model will be a 2D tensor.
            - `max` means that global max pooling will
                be applied.
        classes: optional number of classes to classify images
            into, only to be specified if `include_top` is True, and
            if no `weights` argument is specified.
    # Returns
        A Keras model instance.
    # Raises
        ValueError: in case of invalid argument for `weights`,
            or invalid input shape.
    """
    global backend, layers, models, keras_utils
    backend, layers, models, keras_utils = get_submodules_from_kwargs(kwargs)

    

    if not (weights in {'imagenet', None} or os.path.exists(weights)):
        raise ValueError('The `weights` argument should be either '
                         '`None` (random initialization), `imagenet` '
                         '(pre-training on ImageNet), '
                         'or the path to the weights file to be loaded.')

    if weights == 'imagenet' and include_top and classes != 1000:
        raise ValueError('If using `weights` as `"imagenet"` with `include_top`'
                         ' as true, `classes` should be 1000')

    # Determine proper input shape
    input_shape = _obtain_input_shape(input_shape,
                                      default_size=default_size,
                                      min_size=32,
                                      data_format=backend.image_data_format(),
                                      require_flatten=include_top,
                                      weights=weights)

    if input_tensor is None:
        img_input = layers.Input(shape=input_shape)
    else:
        if not backend.is_keras_tensor(input_tensor):
            img_input = layers.Input(tensor=input_tensor, shape=input_shape)
        else:
            img_input = input_tensor

    bn_axis = 3 if backend.image_data_format() == 'channels_last' else 1

    def round_filters(filters, divisor=depth_divisor):
        """Round number of filters based on depth multiplier."""
        filters *= width_coefficient
        new_filters = max(divisor, int(filters + divisor / 2) // divisor * divisor)
        # Make sure that round down does not go down by more than 10%.
        if new_filters < 0.9 * filters:
            new_filters += divisor
        return int(new_filters)

    def round_repeats(repeats):
        """Round number of repeats based on depth multiplier."""
        return int(math.ceil(depth_coefficient * repeats))

    # Build stem
    x = img_input
    x = layers.ZeroPadding2D(padding=correct_pad(backend, x, 3),
                             name='stem_conv_pad')(x)
    x = layers.Conv2D(round_filters(32), 3,
                      strides=2,
                      padding='valid',
                      use_bias=False,
                      kernel_initializer=CONV_KERNEL_INITIALIZER,
                      name='stem_conv')(x)
    x = layers.BatchNormalization(axis=bn_axis, name='stem_bn')(x)
    x = layers.Activation(activation_fn, name='stem_activation')(x)

    # Build blocks
    from copy import deepcopy
    blocks_args = deepcopy(blocks_args)

    b = 0
    blocks = float(sum(args['repeats'] for args in blocks_args))
    for (i, args) in enumerate(blocks_args):
        assert args['repeats'] > 0
        # Update block input and output filters based on depth multiplier.
        args['filters_in'] = round_filters(args['filters_in'])
        args['filters_out'] = round_filters(args['filters_out'])

        for j in range(round_repeats(args.pop('repeats'))):
            # The first block needs to take care of stride and filter size increase.
            if j > 0:
                args['strides'] = 1
                args['filters_in'] = args['filters_out']
            x = block(x, activation_fn, drop_connect_rate * b / blocks,
                      name='block{}{}_'.format(i + 1, chr(j + 97)), **args)
            b += 1

    # Build top
    x = layers.Conv2D(round_filters(1280), 1,
                      padding='same',
                      use_bias=False,
                      kernel_initializer=CONV_KERNEL_INITIALIZER,
                      name='top_conv')(x)
    x = layers.BatchNormalization(axis=bn_axis, name='top_bn')(x)
    x = layers.Activation(activation_fn, name='top_activation')(x)
    if include_top:
        x = layers.GlobalAveragePooling2D(name='avg_pool')(x)
        if dropout_rate > 0:
            x = layers.Dropout(dropout_rate, name='top_dropout')(x)
        x = layers.Dense(classes,
                         activation='softmax',
                         kernel_initializer=DENSE_KERNEL_INITIALIZER,
                         name='probs')(x)
    else:
        if pooling == 'avg':
            x = layers.GlobalAveragePooling2D(name='avg_pool')(x)
        elif pooling == 'max':
            x = layers.GlobalMaxPooling2D(name='max_pool')(x)

    # Ensure that the model takes into account
    # any potential predecessors of `input_tensor`.
    if input_tensor is not None:
        inputs = keras_utils.get_source_inputs(input_tensor)
    else:
        inputs = img_input

    # Create model.
    model = models.Model(inputs, x, name=model_name)

    # Load weights.
    if weights == 'imagenet':
        if include_top:
            file_suff = '_weights_tf_dim_ordering_tf_kernels_autoaugment.h5'
            file_hash = WEIGHTS_HASHES[model_name[-2:]][0]
        else:
            file_suff = '_weights_tf_dim_ordering_tf_kernels_autoaugment_notop.h5'
            file_hash = WEIGHTS_HASHES[model_name[-2:]][1]
        file_name = model_name + file_suff
        weights_path = keras_utils.get_file(file_name,
                                            BASE_WEIGHTS_PATH + file_name,
                                            cache_subdir='models',
                                            file_hash=file_hash)
        model.load_weights(weights_path)
    elif weights is not None:
        model.load_weights(weights)

    return model


def EfficientNetB0(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(1.0, 1.0, 224, 0.2,
                        model_name='efficientnet-b0',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


def EfficientNetB1(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(1.0, 1.1, 240, 0.2,
                        model_name='efficientnet-b1',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


def EfficientNetB2(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(1.1, 1.2, 260, 0.3,
                        model_name='efficientnet-b2',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


def EfficientNetB3(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(1.2, 1.4, 300, 0.3,
                        model_name='efficientnet-b3',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


def EfficientNetB4(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(1.4, 1.8, 380, 0.4,
                        model_name='efficientnet-b4',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


def EfficientNetB5(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(1.6, 2.2, 456, 0.4,
                        model_name='efficientnet-b5',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


def EfficientNetB6(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(1.8, 2.6, 528, 0.5,
                        model_name='efficientnet-b6',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


def EfficientNetB7(include_top=True,
                   weights='imagenet',
                   input_tensor=None,
                   input_shape=None,
                   pooling=None,
                   classes=1000,
                   **kwargs):
    return EfficientNet(2.0, 3.1, 600, 0.5,
                        model_name='efficientnet-b7',
                        include_top=include_top, weights=weights,
                        input_tensor=input_tensor, input_shape=input_shape,
                        pooling=pooling, classes=classes,
                        **kwargs)


def preprocess_input(x, data_format=None, **kwargs):
    """Preprocesses a numpy array encoding a batch of images.
    # Arguments
        x: a 3D or 4D numpy array consists of RGB values within [0, 255].
        data_format: data format of the image tensor.
    # Returns
        Preprocessed array.
    """
    return tf.keras.applications.imagenet_utils.preprocess_input(x, data_format,
                                           mode='torch', **kwargs)



class KerasEfficientNetB0(AbstractClassificationModel):

    def __init__(self):
        pass

    @classmethod
    def get_base_model(cls,input:(int,int,int),include_top:bool=False, weights:str="imagenet",**kwargs)->tf.keras.Model:

        mish = kwargs.get('mish', True)
        if mish == True:
            activation = Mish
        else:
            activation = swish
        
        return EfficientNetB0(input_shape=input, activation=activation, include_top=include_top, weights=weights)


class KerasEfficientNetB1(AbstractClassificationModel):

    def __init__(self):
        pass

    @classmethod
    def get_base_model(cls,input:(int,int,int),include_top:bool=False, weights:str="imagenet",**kwargs)->tf.keras.Model:

        mish = kwargs.get('mish', True)
        if mish == True:
            activation = Mish
        else:
            activation = swish
        
        return EfficientNetB1(input_shape=input, activation=activation, include_top=include_top, weights=weights)


class KerasEfficientNetB2(AbstractClassificationModel):

    def __init__(self):
        pass

    @classmethod
    def get_base_model(cls,input:(int,int,int),include_top:bool=False, weights:str="imagenet",**kwargs)->tf.keras.Model:

        mish = kwargs.get('mish', True)
        if mish == True:
            activation = Mish
        else:
            activation = swish
        
        return EfficientNetB2(input_shape=input, activation=activation, include_top=include_top, weights=weights)


class KerasEfficientNetB3(AbstractClassificationModel):

    def __init__(self):
        pass

    @classmethod
    def get_base_model(cls,input:(int,int,int),include_top:bool=False, weights:str="imagenet",**kwargs)->tf.keras.Model:

        mish = kwargs.get('mish', True)
        if mish == True:
            activation = Mish
        else:
            activation = swish
        
        return EfficientNetB3(input_shape=input, activation=activation, include_top=include_top, weights=weights)


class KerasEfficientNetB4(AbstractClassificationModel):

    def __init__(self):
        pass

    @classmethod
    def get_base_model(cls,input:(int,int,int),include_top:bool=False, weights:str="imagenet",**kwargs)->tf.keras.Model:
        if include_top == True:
            raise NotImplementedError("This model not support include top")

        mish = kwargs.get('mish', True)
        if mish == True:
            activation = Mish
        else:
            activation = swish
        
        return EfficientNetB4(input_shape=input, activation=activation, include_top=include_top, weights=weights)


class KerasEfficientNetB5(AbstractClassificationModel):

    def __init__(self):
        pass

    @classmethod
    def get_base_model(cls,input:(int,int,int),include_top:bool=False, weights:str="imagenet",**kwargs)->tf.keras.Model:
        if include_top == True:
            raise NotImplementedError("This model not support include top")

        mish = kwargs.get('mish', True)
        if mish == True:
            activation = Mish
        else:
            activation = swish
        
        return EfficientNetB5(input_shape=input, activation=activation, include_top=include_top, weights=weights)

class KerasEfficientNetB6(AbstractClassificationModel):

    def __init__(self):
        pass

    @classmethod
    def get_base_model(cls,input:(int,int,int),include_top:bool=False, weights:str="imagenet",**kwargs)->tf.keras.Model:
        if include_top == True:
            raise NotImplementedError("This model not support include top")

        mish = kwargs.get('mish', True)
        if mish == True:
            activation = Mish
        else:
            activation = swish
        
        return EfficientNetB6(input_shape=input, activation=activation, include_top=include_top, weights=weights)

class KerasEfficientNetB7(AbstractClassificationModel):

    def __init__(self):
        pass

    @classmethod
    def get_base_model(cls,input:(int,int,int),include_top:bool=False, weights:str="imagenet",**kwargs)->tf.keras.Model:

        mish = kwargs.get('mish', True)
        if mish == True:
            activation = Mish
        else:
            activation = swish
        
        return EfficientNetB7(input_shape=input, activation=activation, include_top=include_top, weights=weights)



"""Tensorflow-Keras Implementation of Mish"""

## Import Necessary Modules
import tensorflow as tf

class Mish(tf.keras.layers.Activation):
    '''
    Mish Activation Function.
    .. math::
        mish(x) = x * tanh(softplus(x)) = x * tanh(ln(1 + e^{x}))
    Shape:
        - Input: Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.
        - Output: Same shape as the input.
    Examples:
        >>> X = Activation('Mish', name="conv1_act")(X_input)
    '''

    def __init__(self, activation, **kwargs):
        super(Mish, self).__init__(activation, **kwargs)
        self.__name__ = 'Mish'

def mish(inputs):
    return inputs * tf.math.tanh(tf.math.softplus(inputs))

tf.keras.utils.get_custom_objects().update({'Mish': Mish(mish)})

def USE_MISH_AS_RELU():
    print("USE_MISH_AS_RELU")
    tf.keras.utils.get_custom_objects().update({'relu': Mish(mish)})    



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tftk',
 'tftk.augment_ok',
 'tftk.image',
 'tftk.image.dataset',
 'tftk.image.model',
 'tftk.optimizer',
 'tftk.rl',
 'tftk.train']

package_data = \
{'': ['*']}

install_requires = \
['graphviz>=0.14,<0.15',
 'gym>=0.17.1,<0.18.0',
 'icrawler>=0.6.2,<0.7.0',
 'optuna>=1.2.0,<2.0.0',
 'pillow>=7.0.0,<8.0.0',
 'pydot-ng>=2.0.0,<3.0.0',
 'tensorboard_plugin_profile>=2.2.0,<3.0.0',
 'tensorflow>=2.2.0,<3.0.0',
 'tensorflow_addons>=0.10.0,<0.11.0',
 'tensorflow_datasets>=3.1.0,<4.0.0',
 'tensorflow_probability>=0.9.0,<0.10.0']

setup_kwargs = {
    'name': 'tftk',
    'version': '0.1.3',
    'description': 'Machine Learning Toolkit using TensorFlow',
    'long_description': None,
    'author': 'Naruhide KITADA',
    'author_email': 'kitfactory@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
